from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import SimpleExpression as SimpleExpressionNode
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "ParenthesizedSimpleExpression",
]


class ParenthesizedSimpleExpression(
    NonTerminal[SimpleExpressionNode],
):
    class Meta:

        tags = {
            non_terminal("Term"): 25_000,
        }

    def build_grammar(self) -> str:
        define = self.ebnf.define
        _non_terminal = self.ebnf.non_terminal
        op = self.ebnf.op

        return define(
            op("("),
            _non_terminal("SimpleExpression"),
            op(")"),
        )

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[bool],
    ) -> SimpleExpressionNode:
        return cast(SimpleExpressionNode, tree.children[0])



from enum import IntEnum

from xformula.arch.enumeration import Iota

__all__ = [
    "Placement",
]


__xf__placement_iota__ = Iota()


class Placement(IntEnum):

    UNDEFINED = __xf__placement_iota__()

    __XF__START__ = __xf__placement_iota__()

    LEFT = __xf__placement_iota__()

    RIGHT = __xf__placement_iota__()

    __XF__END__ = __xf__placement_iota__()


del __xf__placement_iota__



from itertools import chain
from typing import Any, Callable, Concatenate, Optional, ParamSpec, TypeVar, cast

from lark import v_args

from xformula.arch.runtime.reflection import ClassBuilder
from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.core.context.abc import SyntaxContext
from xformula.syntax.parser.transformers.abc import ASTBuilderProtocol

__all__ = [
    "ASTBuilderClassBuilder",
]


P = ParamSpec("P")

R = TypeVar("R")


class ASTBuilderClassBuilder(
    ClassBuilder[ASTBuilderProtocol],
):
    @classmethod
    def get_name(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        return "ASTBuilder"

    @classmethod
    def get_bases(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[type, ...]:
        return (ASTBuilderProtocol,)

    @classmethod
    def get_namespace(
        cls,
        syntax_context: SyntaxContext,
        runtime_context: RuntimeContext,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Any]:
        namespace: dict[str, Any] = dict()
        definitions = chain(
            chain(
                syntax_context.non_terminals,
                syntax_context.terminals,
            ),
        )
        for definition in definitions:
            transformer_function: Optional[Callable]
            if isinstance(definition.__class__.options.transform, str):
                transformer_function = getattr(
                    definition,
                    definition.__class__.options.transform,
                    None,
                )
            else:
                transformer_function = definition.__class__.options.transform
            if not callable(transformer_function):
                raise RuntimeError(
                    f"invalid transformer function:"
                    f" {definition.__class__.__qualname__}.{transformer_function!r}"
                    " is not a callable"
                )
            delegate = cls.build_transformer_function_delegator(
                runtime_context,
                definition.__class__.options.definition_name,
                transformer_function,
            )
            namespace[definition.__class__.options.definition_name] = delegate
        return namespace

    @classmethod
    def build_transformer_function_delegator(
        cls,
        runtime_context: RuntimeContext,
        transformer_function_name: str,
        transformer_function: Callable[Concatenate[RuntimeContext, P], R],
    ) -> Callable[Concatenate[ASTBuilderProtocol, P], R]:
        def delegate(
            transformer: ASTBuilderProtocol,
            *args: P.args,
            **kwargs: P.kwargs,
        ) -> R:
            return transformer_function(runtime_context, *args, **kwargs)

        delegate.__name__ = transformer_function_name

        return delegate

    @classmethod
    def build(
        cls,
        syntax_context: SyntaxContext,
        runtime_context: RuntimeContext,
        *args: Any,
        **kwargs: Any,
    ) -> type[ASTBuilderProtocol]:
        built_type = super(ASTBuilderClassBuilder, cls).build(
            syntax_context,
            runtime_context,
            *args,
            **kwargs,
        )
        v_args_kwargs = kwargs.get(
            "v_args_kwargs",
            dict(
                tree=True,
            ),
        )
        if v_args_kwargs:
            built_type = cast(
                type[ASTBuilderProtocol],
                v_args(**v_args_kwargs)(built_type),
            )
        return built_type



from xformula.syntax.core.features.separators.definitions.templates.separated_one_or_more_of import (
    SeparatedOneOrMoreOf,
)
from xformula.syntax.core.features.separators.definitions.templates.separated_pair import (
    SeparatedPair,
)
from xformula.syntax.core.features.separators.definitions.templates.separated_zero_or_more_of import (
    SeparatedZeroOrMoreOf,
)

__all__ = [
    "SeparatedOneOrMoreOf",
    "SeparatedPair",
    "SeparatedZeroOrMoreOf",
]



from functools import partial
from typing import Any, Callable, TypeVar

__all__ = [
    "Bisect",
]


T = TypeVar("T", bound=Any)

V = TypeVar("V", bound=Any)


class Bisect:
    @staticmethod
    def pure(x: T) -> T:
        return x

    @staticmethod
    def left(rhs: V, lhs: V) -> bool:
        return lhs >= rhs

    @staticmethod
    def right(rhs: V, lhs: V) -> bool:
        return lhs > rhs

    @classmethod
    def search(
        cls,
        predict: Callable[[V], bool],
        list_: list[T],
        low: int = 0,
        high: int | None = None,
        key: Callable[[T], V] = pure,
    ) -> int:
        if high is None:
            high = len(list_)
        while low < high:
            middle = (low + high) >> 0x1
            lhs = key(list_[middle])
            if predict(lhs):
                high = middle
            else:
                low = middle + 1
        return low

    @classmethod
    def find_left_index(
        cls,
        list_: list[T],
        x: V,
        low: int = 0,
        high: int | None = None,
        lhs_key: Callable[[T], V] = pure,
        rhs_key: Callable[[T], V] = pure,
    ) -> int:
        rhs = rhs_key(x)
        return cls.search(
            partial(cls.left, rhs),
            list_,
            low=low,
            high=high,
            key=lhs_key,
        )

    @classmethod
    def find_right_index(
        cls,
        list_: list[T],
        x: V,
        low: int = 0,
        high: int | None = None,
        lhs_key: Callable[[T], V] = pure,
        rhs_key: Callable[[T], V] = pure,
    ) -> int:
        rhs = rhs_key(x)
        return cls.search(
            partial(cls.right, rhs),
            list_,
            low=low,
            high=high,
            key=lhs_key,
        )

    @classmethod
    def insert_left(
        cls,
        list_: list[T],
        x: T,
        low: int = 0,
        high: int | None = None,
        lhs_key: Callable[[T], V] = pure,
        rhs_key: Callable[[T], V] = pure,
    ) -> list[T]:
        index = cls.find_left_index(
            list_,
            x,
            low=low,
            high=high,
            lhs_key=lhs_key,
            rhs_key=rhs_key,
        )
        list_.insert(index, x)
        return list_

    @classmethod
    def insert_right(
        cls,
        list_: list[T],
        x: T,
        low: int = 0,
        high: int | None = None,
        lhs_key: Callable[[T], V] = pure,
        rhs_key: Callable[[T], V] = pure,
    ) -> list[T]:
        index = cls.find_right_index(
            list_,
            x,
            low=low,
            high=high,
            lhs_key=lhs_key,
            rhs_key=rhs_key,
        )
        list_.insert(index, x)
        return list_



import dataclasses

from xformula.syntax.ast.nodes import Literal

__all__ = [
    "None_",
]


@dataclasses.dataclass()
class None_(  # NOSONAR
    Literal[None],
):

    value: None = dataclasses.field(
        kw_only=True,
        init=False,
        default=None,
    )



from xformula.syntax.grammar.templates.abc import Template

__all__ = [
    "WrapOneOrMoreOf",
]


class WrapOneOrMoreOf(Template):
    class Meta:

        arguments = [
            Template.arg("left"),
            Template.arg("right"),
            Template.arg("rule"),
        ]

    def build_grammar(self) -> str:
        suite = self.ebnf.suite
        define = self.ebnf.define
        orelse = self.ebnf.orelse
        arg = self.ebnf.arg
        use_template = self.ebnf.use_template
        optional = self.ebnf.optional
        op = self.ebnf.op

        return suite(
            define(
                arg("left"),
                use_template("SeparatedOneOrMoreOf", op(","), arg("rule")),
                optional(op(",")),
                arg("right"),
            ),
            orelse(arg("left"), arg("rule"), optional(op(",")), arg("right")),
        )



from xformula.syntax.grammar.regex._any_char import any_char
from xformula.syntax.grammar.regex._any_char_except import any_char_except
from xformula.syntax.grammar.regex._any_char_of import any_char_of
from xformula.syntax.grammar.regex._any_of import any_of
from xformula.syntax.grammar.regex._bound import bound
from xformula.syntax.grammar.regex._chars_between import chars_between
from xformula.syntax.grammar.regex._escape_char import escape_char
from xformula.syntax.grammar.regex._flat import flat
from xformula.syntax.grammar.regex._flat_join import flat_join
from xformula.syntax.grammar.regex._flat_one_or_more_of import flat_one_or_more_of
from xformula.syntax.grammar.regex._flat_optional import flat_optional
from xformula.syntax.grammar.regex._flat_zero_or_more_of import flat_zero_or_more_of
from xformula.syntax.grammar.regex._non_capturing_group import non_capturing_group
from xformula.syntax.grammar.regex._one_or_more_of import one_or_more_of
from xformula.syntax.grammar.regex._whitespace import whitespace
from xformula.syntax.grammar.regex._word import word
from xformula.syntax.grammar.regex._wrap import wrap
from xformula.syntax.grammar.regex._zero_or_more_of import zero_or_more_of

__all__ = [
    "any_char",
    "any_char_except",
    "any_char_of",
    "any_of",
    "bound",
    "chars_between",
    "escape_char",
    "flat",
    "flat_join",
    "flat_one_or_more_of",
    "flat_optional",
    "flat_zero_or_more_of",
    "non_capturing_group",
    "one_or_more_of",
    "whitespace",
    "word",
    "wrap",
    "zero_or_more_of",
]



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "IdentifierFeature",
]


class IdentifierFeature(Feature):
    class Meta:
        fqn = "xformula.core.Identifier"

    def setup(self) -> None:
        from xformula.syntax.core.features.identifiers.definitions.non_terminals import (
            Identifier,
        )
        from xformula.syntax.core.features.identifiers.definitions.terminals import (
            IDENTIFIER,
        )

        self.non_terminal_types.append(Identifier)
        self.terminal_types.append(IDENTIFIER)



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import SimpleExpression
from xformula.syntax.core.features.mappings.ast.nodes import DictPair as DictPairNode
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "DictPair",
]


class DictPair(
    NonTerminal[DictPairNode],
):
    class Meta:

        atomic = True

    def build_grammar(self) -> str:
        define = self.ebnf.define
        use_template = self.ebnf.use_template
        op = self.ebnf.op
        _non_terminal = self.ebnf.non_terminal

        return define(
            use_template(
                "SeparatedPair",
                op(":"),
                _non_terminal("SimpleExpression"),
                _non_terminal("SimpleExpression"),
            ),
        )

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[SimpleExpression],
    ) -> DictPairNode:
        elements = cast(
            tuple[SimpleExpression, SimpleExpression],
            tuple(tree.children),
        )
        return DictPairNode(
            elements=elements,
        )



from xformula.syntax.grammar.ebnf._flat_join import flat_join

__all__ = [
    "group",
]


def group(*expressions: str) -> str:
    inner = flat_join(" ", expressions)
    if not inner:
        return ""
    return f"({inner})"



__all__ = [
    "MissingDefinitionError",
]


class MissingDefinitionError(Exception):

    missing_definitions: dict[str, dict[str, set[str]]]

    @classmethod
    def build_message(
        cls,
        missing_definitions: dict[str, dict[str, set[str]]],
    ) -> str:
        message = "\n".join(
            [
                f"Missing {definition_type} definitions:\n"
                f"\t⧃ {missing_name} : ({' | '.join(alternation)})"
                for definition_type, detected in missing_definitions.items()
                for missing_name, alternation in detected.items()
            ],
        )
        return message

    def __init__(
        self,
        missing_definitions: dict[str, dict[str, set[str]]],
    ) -> None:
        self.missing_definitions = missing_definitions
        super(MissingDefinitionError, self).__init__(
            f"\n\n{self.__class__.build_message(self.missing_definitions)}",
        )



import dataclasses

from xformula.syntax.ast.nodes.abc import (
    Arity,
    Associativity,
    Placement,
    PrecedenceLevel,
    SymbolType,
)

__all__ = [
    "OperatorPrecedence",
]


@dataclasses.dataclass()
class OperatorPrecedence:

    arity: Arity = dataclasses.field(
        kw_only=True,
        default=0,
    )

    associativity: Associativity = dataclasses.field(
        kw_only=True,
        default=Associativity.UNDEFINED,
    )

    level: PrecedenceLevel = dataclasses.field(
        kw_only=True,
        default=int(),
    )

    placement: Placement = dataclasses.field(
        kw_only=True,
        default=Placement.UNDEFINED,
    )

    symbols: list[tuple[SymbolType, str]] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )



from xformula.syntax.grammar.ebnf._template import template

__all__ = [
    "use_template",
]


def use_template(
    name: str,
    *arguments: str,
) -> str:
    args = ", ".join(arguments)
    specify_args = "".join(["{", args, "}"])
    return f"{template(name)}{specify_args}"



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import Term as TermNode
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "Term",
]


class Term(
    NonTerminal[TermNode],
):
    class Meta:

        tags = {
            non_terminal("Primary"): 0,
        }

    def build_grammar(self) -> str:
        return self.ebnf.define_tagged_alternation()

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[bool],
    ) -> TermNode:
        return cast(TermNode, tree.children[0])



from typing import TypeVar

from xformula.syntax.ast.nodes.abc.has_value import HasValue
from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.term import Term

__all__ = [
    "Literal",
]


T = TypeVar("T")


class Literal(
    Term,
    HasValue[T],
    Node,
):
    class Meta(
        Term.Meta,
        HasValue.Meta,
        Node.Meta,
    ):

        abstract = True

        is_literal = True



import dataclasses

from xformula.syntax.ast.nodes import Literal

__all__ = [
    "Int",
]


@dataclasses.dataclass()
class Int(
    Literal[int],
):

    value: int = dataclasses.field(
        kw_only=True,
        default=int(),
    )



from abc import ABC
from typing import Generic, TypeVar

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.grammar.definitions.abc import Definition, DefinitionType
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "NonTerminal",
]


T = TypeVar("T")


class NonTerminal(
    Definition,
    ABC,
    Generic[T],
):
    class Meta(Definition.Meta):

        abstract = True

        type = DefinitionType.NON_TERMINAL

        transform = "transform_parse_tree"

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree,
    ) -> T:
        raise NotImplementedError(
            f"{self.__class__.__qualname__}.transform_parse_tree is not implemented",
        )



from typing import TypeVar

from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.simple_expression import SimpleExpression

__all__ = [
    "Operand",
]


N = TypeVar("N", bound=Node)


class Operand(
    SimpleExpression,
    Node,
):
    class Meta(
        SimpleExpression.Meta,
        Node.Meta,
    ):

        abstract = True

        is_operand = True



from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.terminals.abc import Terminal
from xformula.syntax.lexer.tokens.abc import Token

__all__ = [
    "SINGLE_LINE_STR",
]


class SINGLE_LINE_STR(  # NOSONAR
    Terminal[str],
):
    class Meta:

        tags = {
            non_terminal("Str"): 1000,
        }

    def build_single_line_quoted_str_regex_pattern(
        self,
        quote: str,
    ) -> str:
        any_char = self.regex.any_char
        any_char_except = self.regex.any_char_except
        any_of = self.regex.any_of
        escape_char = self.regex.escape_char
        flat_zero_or_more_of = self.regex.flat_zero_or_more_of
        non_capturing_group = self.regex.non_capturing_group
        wrap = self.regex.wrap

        return wrap(
            escape_char(quote),
            flat_zero_or_more_of(
                non_capturing_group(
                    any_of(
                        non_capturing_group(
                            escape_char(r"\\"),
                            any_char(),
                        ),
                        non_capturing_group(
                            any_char_except(
                                escape_char(quote),
                                r"\n",
                                escape_char(r"\\"),
                            ),
                        ),
                    ),
                ),
            ),
        )

    def build_grammar(self) -> str:
        define = self.ebnf.define
        regex = self.ebnf.regex

        any_of = self.regex.any_of

        re_single_line_single_quoted = self.build_single_line_quoted_str_regex_pattern(
            "'"
        )
        re_single_line_double_quoted = self.build_single_line_quoted_str_regex_pattern(
            '"'
        )

        return define(
            regex(
                any_of(
                    re_single_line_single_quoted,
                    re_single_line_double_quoted,
                ),
            ),
        )

    def transform_token(
        self,
        runtime_context: RuntimeContext,
        token: Token,
    ) -> str:
        return bytes(token.value[1:-1], "unicode_escape").decode("unicode_escape")



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from xformula.syntax.core.context.abc.syntax_context import SyntaxContext

__all__ = [
    "MissingNonTerminalDetector",
]


class MissingNonTerminalDetector:

    syntax_context: "SyntaxContext"

    @classmethod
    def detect(
        cls,
        syntax_context: "SyntaxContext",
    ) -> dict[str, set[str]]:
        detector = cls(syntax_context)
        return detector.find_tags_have_no_definition()

    def __init__(self, syntax_context: "SyntaxContext") -> None:
        self.syntax_context = syntax_context

    def find_tags_have_no_definition(self) -> dict[str, set[str]]:
        detected: dict[str, set[str]] = dict()
        defined_tags = {
            definition.__class__.options.definition_name
            for definition in self.syntax_context.non_terminals
        }
        for tag, alternation in self.syntax_context.tagged_definitions.items():
            if tag not in defined_tags:
                detected[tag] = {
                    wanted_by_definition.__class__.options.definition_name
                    for wanted_by_definition in alternation
                }
        return detected



from xformula.syntax.grammar.ebnf._any_of import any_of
from xformula.syntax.grammar.ebnf._arg import arg
from xformula.syntax.grammar.ebnf._concat import concat
from xformula.syntax.grammar.ebnf._define import define
from xformula.syntax.grammar.ebnf._define_rule import define_rule
from xformula.syntax.grammar.ebnf._directive import directive
from xformula.syntax.grammar.ebnf._document import document
from xformula.syntax.grammar.ebnf._flat_join import flat_join
from xformula.syntax.grammar.ebnf._get_normalized_name_by_definition import (
    get_normalized_name_by_definition,
)
from xformula.syntax.grammar.ebnf._group import group
from xformula.syntax.grammar.ebnf._indent import indent
from xformula.syntax.grammar.ebnf._keyword import keyword
from xformula.syntax.grammar.ebnf._literal import literal
from xformula.syntax.grammar.ebnf._non_terminal import non_terminal
from xformula.syntax.grammar.ebnf._normalize_name_by_type import normalize_name_by_type
from xformula.syntax.grammar.ebnf._one_or_more_of import one_or_more_of
from xformula.syntax.grammar.ebnf._op import op
from xformula.syntax.grammar.ebnf._optional import optional
from xformula.syntax.grammar.ebnf._orelse import orelse
from xformula.syntax.grammar.ebnf._regex import regex
from xformula.syntax.grammar.ebnf._suite import suite
from xformula.syntax.grammar.ebnf._sym import sym
from xformula.syntax.grammar.ebnf._template import template
from xformula.syntax.grammar.ebnf._terminal import terminal
from xformula.syntax.grammar.ebnf._use_directive import use_directive
from xformula.syntax.grammar.ebnf._use_template import use_template

__all__ = [
    "any_of",
    "arg",
    "concat",
    "define",
    "define_rule",
    "directive",
    "document",
    "flat_join",
    "get_normalized_name_by_definition",
    "group",
    "indent",
    "keyword",
    "literal",
    "non_terminal",
    "normalize_name_by_type",
    "one_or_more_of",
    "op",
    "optional",
    "orelse",
    "regex",
    "suite",
    "sym",
    "template",
    "terminal",
    "use_directive",
    "use_template",
]



from typing import TypeVar

from xformula.syntax.ast.nodes.abc.container import Container
from xformula.syntax.ast.nodes.abc.node import Node

__all__ = [
    "List",
]


T = TypeVar("T")


class List(
    Container[T],
    Node,
):
    class Meta(
        Container.Meta,
        Node.Meta,
    ):

        abstract = True

        is_list = True



import dataclasses

from xformula.syntax.ast.nodes.abc import Dict as _Dict
from xformula.syntax.ast.nodes.abc import Pair, SimpleExpression

__all__ = [
    "Dict",
]


@dataclasses.dataclass()
class Dict(
    _Dict[SimpleExpression, SimpleExpression],
):

    elements: list[Pair[SimpleExpression, SimpleExpression]] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )



from abc import ABC
from typing import Any, Generic, TypeVar, cast

__all__ = [
    "ClassBuilder",
]


T = TypeVar("T")


class ClassBuilder(
    Generic[T],
    ABC,
):
    @classmethod
    def get_name(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        raise NotImplementedError(f"{cls.__qualname__}.get_name is not implemented")

    @classmethod
    def get_base_qualname(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        return cls.__qualname__

    @classmethod
    def get_bases(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[type, ...]:
        return ()

    @classmethod
    def get_namespace(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Any]:
        return dict()

    @classmethod
    def build(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> type[T]:
        name = cls.get_name(*args, **kwargs)
        base_qualname = cls.get_base_qualname(*args, **kwargs)
        bases = cls.get_bases(*args, **kwargs)
        namespace = cls.get_namespace(*args, **kwargs)
        built_type = cast(type[T], type(name, bases, namespace))
        built_type.__qualname__ = ".".join([base_qualname, name])
        return built_type



from typing import TYPE_CHECKING

from xformula.syntax.core.features.polyfill.runtime.reflection import (
    NonTerminalClassBuilder,
)
from xformula.syntax.grammar.non_terminals.abc import NonTerminal

if TYPE_CHECKING:
    from xformula.syntax.core.context.abc.syntax_context import SyntaxContext

__all__ = [
    "NonTerminalTypeGenerator",
]


class NonTerminalTypeGenerator:

    syntax_context: "SyntaxContext"

    missing_definitions: dict[str, set[str]]

    @classmethod
    def generate(
        cls,
        syntax_context: "SyntaxContext",
        missing_definitions: dict[str, set[str]],
    ) -> list[type[NonTerminal]]:
        generator = cls(syntax_context, missing_definitions)
        return generator.generate_types()

    def __init__(
        self,
        syntax_context: "SyntaxContext",
        missing_definitions: dict[str, set[str]],
    ) -> None:
        self.syntax_context = syntax_context
        self.missing_definitions = missing_definitions

    def generate_types(self) -> list[type[NonTerminal]]:
        generated_types: list[type[NonTerminal]] = []
        for name, alternation in self.missing_definitions.items():
            generated_type = NonTerminalClassBuilder.build(name, alternation)
            generated_types.append(generated_type)
        return generated_types



import dataclasses

from xformula.arch.meta import Options

__all__ = [
    "NodeOptions",
]


@dataclasses.dataclass()
class NodeOptions(Options):

    is_start_point: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_expression: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_simple_expression: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_operation: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_operator: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_symbol: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_operand: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_primary: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_mapping: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_dict: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_pair: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_container: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_set: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_tuple: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_list: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_term: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_call: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_identifier: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    is_literal: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_associativity_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_arity_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_arguments_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_callee_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_context_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_elements_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_name_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_owner_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_placement_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_precedence_level_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_type_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_operands_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_operator_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_symbols_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )

    has_value_attribute: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )



from xformula.syntax.core.features.wrappers.definitions.templates.wrap_one_or_more_of import (
    WrapOneOrMoreOf,
)
from xformula.syntax.core.features.wrappers.definitions.templates.wrap_zero_or_more_of import (
    WrapZeroOrMoreOf,
)

__all__ = [
    "WrapOneOrMoreOf",
    "WrapZeroOrMoreOf",
]



__all__ = [
    "indent",
]


def indent(
    *block: str,
    width: int = 4,
) -> str:
    width = max(4, width)

    tabs = "".join([(" " * 4) for _ in range(width // 4)])
    spaces = "".join([" " for _ in range(width % 4)])

    prefix = f"{tabs}{spaces}"

    return "\n".join(
        [
            f"{prefix}{line}".rstrip()
            for statement in block
            for line in statement.split("\n")
        ],
    )



from xformula.syntax.grammar.regex._flat_join import flat_join

__all__ = [
    "bound",
]


def bound(*expressions):
    inner = flat_join("", expressions)
    if not inner:
        return ""
    return rf"\b{inner}\b"



from xformula.syntax.grammar.directives.abc import Directive

__all__ = [
    "SkipWhitespace",
]


class SkipWhitespace(Directive):
    def build_grammar(self) -> str:
        use_directive = self.ebnf.use_directive
        terminal = self.ebnf.terminal

        return use_directive("ignore", terminal("WHITESPACE"))



from typing import Any, NoReturn, TypeVar

from xformula.arch.meta import Meta
from xformula.syntax.grammar.definitions.abc.definition_options import DefinitionOptions

__all__ = [
    "DefinitionMeta",
]


C = TypeVar("C", bound="DefinitionMeta")


class DefinitionMeta(
    Meta[DefinitionOptions],
):

    options_class = DefinitionOptions

    options: DefinitionOptions

    @classmethod
    def get_excluded_super_options_attnames(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> set[str]:
        excluded = super(DefinitionMeta, mcs).get_excluded_super_options_attnames(
            name,
            bases,
            namespace,
            **kwargs,
        )
        excluded.add("definition_name")
        return excluded

    @classmethod
    def get_super_options_kwargs(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> dict[str, Any]:
        options_kwargs = super(DefinitionMeta, mcs).get_super_options_kwargs(
            name,
            bases,
            namespace,
            **kwargs,
        )
        if "arguments" in options_kwargs:
            options_kwargs["arguments"] = options_kwargs["arguments"].copy()
        if "tags" in options_kwargs:
            options_kwargs["tags"] = options_kwargs["tags"].copy()
        return options_kwargs

    @classmethod
    def populate_options_kwargs(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        meta: Any | None,
        options_kwargs: dict[str, Any],
        **kwargs: Any,
    ) -> None:
        from xformula.syntax.grammar.ebnf import normalize_name_by_type

        super(DefinitionMeta, mcs).populate_options_kwargs(
            name,
            bases,
            namespace,
            meta,
            options_kwargs,
            **kwargs,
        )

        abstract = options_kwargs.get("abstract", False)

        if not abstract:

            if not options_kwargs.get("definition_name", ""):
                options_kwargs["definition_name"] = options_kwargs["name"]

            if not abstract and "type" in options_kwargs:
                options_kwargs["definition_name"] = normalize_name_by_type(
                    options_kwargs["type"],
                    options_kwargs["definition_name"],
                )

    @classmethod
    def validate_options(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        options: DefinitionOptions,
        **kwargs: Any,
    ) -> NoReturn | None:
        super(DefinitionMeta, mcs).validate_options(
            name,
            bases,
            namespace,
            options,
            **kwargs,
        )

        if options.abstract:
            return None

        from xformula.syntax.grammar.definitions.abc.definition_type import (
            DefinitionType,
        )

        if not (
            DefinitionType.__XF__START__ < options.type < DefinitionType.__XF__END__
        ):
            raise TypeError(f"invalid rule type: {options.type!r}")

        return None



from typing import TypeVar

from xformula.syntax.ast.nodes.abc.mapping import Mapping
from xformula.syntax.ast.nodes.abc.node import Node

__all__ = [
    "Dict",
]


K = TypeVar("K")

V = TypeVar("V")


class Dict(
    Mapping[K, V],
    Node,
):
    class Meta(
        Mapping.Meta,
        Node.Meta,
    ):

        abstract = True

        is_dict = True



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "OperationFeature",
]


class OperationFeature(Feature):
    class Meta:
        fqn = "xformula.core.Operation"

    def setup(self) -> None:
        from xformula.syntax.core.features.operations.definitions.non_terminals import (
            Operand,
            Operation,
        )
        from xformula.syntax.core.features.operations.generators import (
            NonTerminalOperationTypeGenerator,
        )

        non_terminal_operation_types = NonTerminalOperationTypeGenerator.generate(
            self.context,
        )

        self.non_terminal_types.append(Operand)
        self.non_terminal_types.extend(non_terminal_operation_types)
        self.non_terminal_types.append(Operation)



from enum import IntEnum

from xformula.arch.enumeration import Iota

__all__ = [
    "SymbolType",
]


__xf__symbol_type_iota__ = Iota()


class SymbolType(IntEnum):

    UNDEFINED = __xf__symbol_type_iota__()

    __XF__START__ = __xf__symbol_type_iota__()

    KEYWORD = __xf__symbol_type_iota__()

    OPERATOR = __xf__symbol_type_iota__()

    __XF__END__ = __xf__symbol_type_iota__()


del __xf__symbol_type_iota__



from xformula.syntax.core.context import SyntaxContext
from xformula.syntax.core.customization.tagging import TaggedDefinitionIterator
from xformula.syntax.core.features import (
    DEFAULT_FEATURE_TYPES,
    Feature,
    FeatureMeta,
    FeatureOptions,
)
from xformula.syntax.core.operations import (
    DEFAULT_OPERATOR_PRECEDENCES,
    OperatorPrecedence,
)

__all__ = [
    "DEFAULT_FEATURE_TYPES",
    "DEFAULT_OPERATOR_PRECEDENCES",
    "Feature",
    "FeatureMeta",
    "FeatureOptions",
    "OperatorPrecedence",
    "SyntaxContext",
    "TaggedDefinitionIterator",
]



from xformula.syntax.grammar.definitions.runtime.reflection.ebnf_expression_builder_class_builder import (
    EBNFExpressionBuilderClassBuilder,
)
from xformula.syntax.grammar.definitions.runtime.reflection.regex_expression_builder_class_builder import (
    RegexExpressionBuilderClassBuilder,
)

__all__ = [
    "EBNFExpressionBuilderClassBuilder",
    "RegexExpressionBuilderClassBuilder",
]



from xformula.syntax.ast.nodes.abc.has_type import HasType
from xformula.syntax.ast.nodes.abc.has_value import HasValue
from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.symbol_type import SymbolType

__all__ = [
    "Symbol",
]


class Symbol(
    HasType[SymbolType],
    HasValue[str],
    Node,
):
    class Meta(
        HasType.Meta,
        HasValue.Meta,
        Node.Meta,
    ):
        abstract = True

        is_symbol = True



from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.precedence_level import PrecedenceLevel

__all__ = [
    "HasPrecedenceLevel",
]


class HasPrecedenceLevel(Node):

    precedence_level: PrecedenceLevel

    class Meta(Node.Meta):

        abstract = True

        has_precedence_level_attribute = True



from enum import IntEnum

from xformula.arch.enumeration import Iota

__all__ = [
    "Context",
]


__xf__context_iota__ = Iota()


class Context(IntEnum):

    UNDEFINED = __xf__context_iota__()

    __XF__START__ = __xf__context_iota__()

    ATTRIBUTE = __xf__context_iota__()

    DELETE = __xf__context_iota__()

    KEYWORD_ARGUMENT = __xf__context_iota__()

    LOAD = __xf__context_iota__()

    REGISTER = __xf__context_iota__()

    __XF__END__ = __xf__context_iota__()


del __xf__context_iota__



import dataclasses

from xformula.syntax.ast.nodes.abc import Arity, Associativity
from xformula.syntax.ast.nodes.abc import Operator as _Operator
from xformula.syntax.ast.nodes.abc import Placement, PrecedenceLevel, Symbol

__all__ = [
    "Operator",
]


@dataclasses.dataclass()
class Operator(_Operator):

    arity: Arity = dataclasses.field(
        kw_only=True,
        default=-1,
    )

    associativity: Associativity = dataclasses.field(
        kw_only=True,
        default=Associativity.UNDEFINED,
    )

    placement: Placement = dataclasses.field(
        kw_only=True,
        default=Placement.UNDEFINED,
    )

    precedence_level: PrecedenceLevel = dataclasses.field(
        kw_only=True,
        default=int(),
    )

    symbols: list[Symbol] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )



from xformula.syntax.ast.nodes.abc.has_operands import HasOperands
from xformula.syntax.ast.nodes.abc.has_operator import HasOperator
from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.simple_expression import SimpleExpression

__all__ = [
    "Operation",
]


class Operation(
    SimpleExpression,
    HasOperator,
    HasOperands,
    Node,
):
    class Meta(
        SimpleExpression.Meta,
        HasOperator.Meta,
        HasOperands.Meta,
        Node.Meta,
    ):

        abstract = True

        is_operation = True



from contextlib import suppress
from typing import TYPE_CHECKING, Iterable, Protocol, Union, runtime_checkable

if TYPE_CHECKING:
    from xformula.syntax.core.context.abc.syntax_context import SyntaxContext
    from xformula.syntax.core.symbols.abc.symbol_type import SymbolType
    from xformula.syntax.grammar.definitions.abc.definition import Definition
    from xformula.syntax.grammar.definitions.abc.definition_type import DefinitionType

__all__ = [
    "EBNFExpressionBuilderProtocol",
]


@runtime_checkable
class EBNFExpressionBuilderProtocol(Protocol):

    _definition: "Definition"

    def __init__(
        self,
        definition: "Definition",
    ) -> None:
        self._definition = definition

    def any_of(
        self,
        *expressions: str,
    ) -> str:
        ...

    def arg(
        self,
        name: str,
    ) -> str:
        ...

    def concat(
        self,
        expressions: Iterable[str],
    ) -> str:
        ...

    def define(
        self,
        *expressions: str,
    ) -> str:
        ...

    def define_rule(
        self,
        definition: "Definition",
    ) -> str:
        ...

    def directive(
        self,
        name: str,
    ) -> str:
        ...

    def document(
        self,
        syntax_context: "SyntaxContext",
    ) -> str:
        ...

    def flat_join(
        self,
        separator: str,
        expressions: Iterable[str],
    ) -> str:
        ...

    def get_normalized_name_by_definition(
        self,
        definition: Union[
            "Definition",
            type["Definition"],
        ],
    ) -> str:
        ...

    def group(
        self,
        *expressions: str,
    ) -> str:
        ...

    def indent(
        self,
        *block: str,
        width: int = 4,
    ) -> str:
        ...

    def keyword(
        self,
        name: str,
    ) -> str:
        ...

    def literal(
        self,
        value: str,
    ) -> str:
        ...

    def non_terminal(
        self,
        name: str,
    ) -> str:
        ...

    def normalize_name_by_type(
        self,
        name: str,
        definition_type: "DefinitionType",
    ) -> str:
        ...

    def one_or_more_of(
        self,
        *expressions: str,
    ) -> str:
        ...

    def op(
        self,
        symbol: str,
    ) -> str:
        ...

    def optional(
        self,
        *expressions: str,
    ) -> str:
        ...

    def orelse(
        self,
        *expressions: str,
    ) -> str:
        ...

    def regex(
        self,
        *expressions: str,
    ) -> str:
        ...

    def suite(
        self,
        *lines: str,
    ) -> str:
        ...

    def sym(
        self,
        info: tuple["SymbolType", str],
    ) -> str:
        ...

    def define_tagged_alternation(self) -> str:
        from xformula.syntax.core.customization import TaggedDefinitionIterator

        alternation = TaggedDefinitionIterator.iterate(
            self._definition.context,
            self._definition.__class__.options.definition_name,
        )

        branch_iterator = iter(alternation)

        first_branch = ""
        with suppress(StopIteration):
            first_branch = next(branch_iterator).__class__.options.definition_name

        define = self.define
        orelse = self.orelse

        return self.suite(
            define(first_branch) if first_branch else "",
            *[
                orelse(branch.__class__.options.definition_name)
                for branch in branch_iterator
            ],
        )

    def tagged_alternation(self) -> str:
        from xformula.syntax.core.customization import TaggedDefinitionIterator

        alternation = TaggedDefinitionIterator.iterate(
            self._definition.context,
            self._definition.__class__.options.definition_name,
        )

        return self.any_of(
            *[branch.__class__.options.definition_name for branch in alternation],
        )

    def or_tagged_alternation(self) -> str:
        str_tagged_alternation = self.tagged_alternation()
        if not str_tagged_alternation:
            return ""

        return self.orelse(str_tagged_alternation)

    def template(
        self,
        name: str,
    ) -> str:
        ...

    def terminal(
        self,
        name: str,
    ) -> str:
        ...

    def use_directive(
        self,
        name: str,
        *expressions: str,
    ) -> str:
        ...

    def use_template(
        self,
        name: str,
        *arguments: str,
    ) -> str:
        ...



from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.symbol import Symbol

__all__ = [
    "HasSymbols",
]


class HasSymbols(Node):

    symbols: list[Symbol]

    class Meta(Node.Meta):

        abstract = True

        has_symbols_attribute = True



from typing import TYPE_CHECKING, Any

from xformula.syntax.core.features.operations.runtime.reflection import (
    NonTerminalOperationClassBuilder,
)
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal

if TYPE_CHECKING:
    from xformula.syntax.core.context.abc.syntax_context import SyntaxContext

__all__ = [
    "NonTerminalOperationTypeGenerator",
]


class NonTerminalOperationTypeGenerator:
    @classmethod
    def generate(
        cls,
        syntax_context: "SyntaxContext",
    ) -> list[type[NonTerminal]]:
        generated_non_terminal_types: list[type[NonTerminal]] = []

        operand = non_terminal("Operand")

        for (
            operator_precedence_index,
            operator_precedence,
        ) in enumerate(syntax_context.operator_precedences):

            kwargs: dict[str, Any] = dict()

            if (
                operator_precedence_index
                == len(syntax_context.operator_precedences) - 1
            ):
                kwargs["tags"] = {
                    non_terminal("Operation"): 0,
                }

            generated_non_terminal_type = NonTerminalOperationClassBuilder.build(
                f"OperationLevel{operator_precedence_index + 1}",
                operator_precedence,
                operand,
                **kwargs,
            )

            generated_non_terminal_types.append(generated_non_terminal_type)
            operand = generated_non_terminal_type.options.definition_name

        return generated_non_terminal_types



import dataclasses

from xformula.syntax.ast.nodes.abc import SimpleExpression
from xformula.syntax.ast.nodes.abc import Tuple as _Tuple

__all__ = [
    "Tuple",
]


@dataclasses.dataclass()
class Tuple(
    _Tuple[SimpleExpression],
):

    elements: list[SimpleExpression] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )



from xformula.syntax.ast.nodes.abc.expression import Expression
from xformula.syntax.ast.nodes.abc.node import Node

__all__ = [
    "SimpleExpression",
]


class SimpleExpression(
    Expression,
    Node,
):
    class Meta(
        Expression.Meta,
        Node.Meta,
    ):

        abstract = True

        is_simple_expression = True



from collections import deque
from enum import IntEnum
from itertools import zip_longest
from typing import Any, Callable, Iterable, Literal, Optional, TypeVar, overload

from xformula.arch.enumeration.iota import Iota

__all__ = [
    "Zip",
]


S = TypeVar("S")

T = TypeVar("T")

V = TypeVar("V", bound=Optional[Any])

X = TypeVar("X")


__xf__zip_separator_placement_iota__ = Iota()


class Zip:
    class SeparatorPlacement(IntEnum):

        UNDEFINED = __xf__zip_separator_placement_iota__()

        __XF__START__ = __xf__zip_separator_placement_iota__()

        LEFT = __xf__zip_separator_placement_iota__()

        RIGHT = __xf__zip_separator_placement_iota__()

        __XF__END__ = __xf__zip_separator_placement_iota__()

    @classmethod
    @overload
    def join(
        cls,
        separators: Iterable[S],
        elements: Iterable[T],
        separator_placement: Literal[SeparatorPlacement.LEFT] = SeparatorPlacement.LEFT,
    ) -> Iterable[tuple[S | None, T | None]]:
        ...

    @classmethod
    @overload
    def join(
        cls,
        separators: Iterable[S],
        elements: Iterable[T],
        separator_placement: Literal[SeparatorPlacement.LEFT] = SeparatorPlacement.LEFT,
        default_value: V = None,
    ) -> Iterable[tuple[S | V, T | V]]:
        ...

    @classmethod
    @overload
    def join(
        cls,
        separators: Iterable[S],
        elements: Iterable[T],
        separator_placement: Literal[
            SeparatorPlacement.RIGHT
        ] = SeparatorPlacement.RIGHT,
    ) -> Iterable[tuple[T | None, S | None]]:
        ...

    @classmethod
    @overload
    def join(
        cls,
        separators: Iterable[S],
        elements: Iterable[T],
        separator_placement: Literal[
            SeparatorPlacement.RIGHT
        ] = SeparatorPlacement.RIGHT,
        default_value: V = None,
    ) -> Iterable[tuple[T | V, S | V]]:
        ...

    @classmethod
    def join(
        cls,
        separators,
        elements,
        separator_placement=SeparatorPlacement.RIGHT,
        default_value=None,
    ):
        if not (
            cls.SeparatorPlacement.__XF__START__
            < separator_placement
            < cls.SeparatorPlacement.__XF__END__
        ):
            raise TypeError(f"invalid separator placement: {separator_placement!r}")

        iterator = zip_longest(
            separators,
            elements,
            fillvalue=default_value,
        )

        for s, x in iterator:
            if separator_placement == cls.SeparatorPlacement.LEFT:
                yield s, x
            elif separator_placement == cls.SeparatorPlacement.RIGHT:
                yield x, s
            else:
                raise NotImplementedError(
                    f"{cls.__qualname__}.join is not implemented for"
                    f" separator placement: {separator_placement!r}"
                )

    @classmethod
    @overload
    def split(
        cls,
        iterable: Iterable[T | S],
    ) -> tuple[Iterable[S], Iterable[T]]:
        ...

    @classmethod
    @overload
    def split(
        cls,
        iterable: Iterable[T | S],
        separator_placement: SeparatorPlacement,
    ) -> tuple[Iterable[S], Iterable[T]]:
        ...

    @classmethod
    @overload
    def split(
        cls,
        iterable: Iterable[S | T],
        separator_placement: Literal[SeparatorPlacement.LEFT] = SeparatorPlacement.LEFT,
    ) -> tuple[Iterable[S], Iterable[T]]:
        ...

    @classmethod
    @overload
    def split(
        cls,
        iterable: Iterable[T | S],
        separator_placement: Literal[
            SeparatorPlacement.RIGHT
        ] = SeparatorPlacement.RIGHT,
    ) -> tuple[Iterable[S], Iterable[T]]:
        ...

    @classmethod
    def split(
        cls,
        iterable,
        separator_placement=SeparatorPlacement.RIGHT,
    ):
        if not (
            cls.SeparatorPlacement.__XF__START__
            < separator_placement
            < cls.SeparatorPlacement.__XF__END__
        ):
            raise TypeError(f"invalid separator placement: {separator_placement!r}")

        iterator = iter(iterable)

        separator_channel: deque[S] = deque()
        element_channel: deque[T] = deque()

        def filter_channel(
            channel: deque[X],
            is_separator: Callable[[int], bool],
        ) -> Iterable[X]:
            nonlocal separator_channel
            nonlocal element_channel

            i = 0
            while True:
                if not channel:
                    try:
                        x = next(iterator)
                    except StopIteration:
                        return

                    if is_separator(i):
                        separator_channel.append(x)
                    else:
                        element_channel.append(x)

                if channel:
                    yield channel.popleft()
                i += 1

        if separator_placement == cls.SeparatorPlacement.LEFT:
            return (
                filter_channel(
                    separator_channel,
                    lambda j: j % 2 == 0,
                ),
                filter_channel(
                    element_channel,
                    lambda j: j % 2 != 0,
                ),
            )

        elif separator_placement == cls.SeparatorPlacement.RIGHT:
            return (
                filter_channel(
                    separator_channel,
                    lambda j: j % 2 != 0,
                ),
                filter_channel(
                    element_channel,
                    lambda j: j % 2 == 0,
                ),
            )

        else:
            raise NotImplementedError(
                f"{cls.__qualname__}.split is not implemented for"
                f" separator placement: {separator_placement!r}"
            )


del __xf__zip_separator_placement_iota__



import dataclasses

from xformula.syntax.ast.nodes.abc import Attribute as _Attribute
from xformula.syntax.ast.nodes.abc import Context, Identifier, Node

__all__ = [
    "Attribute",
]


@dataclasses.dataclass()
class Attribute(_Attribute):

    owner: Node = dataclasses.field(
        kw_only=True,
        default_factory=Node,
    )

    name: Identifier = dataclasses.field(
        kw_only=True,
        default_factory=Identifier,
    )

    context: Context = dataclasses.field(
        kw_only=True,
        default=Context.UNDEFINED,
    )



import dataclasses

from xformula.syntax.ast.nodes.abc import Context
from xformula.syntax.ast.nodes.abc import Identifier as _Identifier

__all__ = [
    "Identifier",
]


@dataclasses.dataclass()
class Identifier(_Identifier):

    name: str = dataclasses.field(
        kw_only=True,
        default=str(),
    )

    context: Context = dataclasses.field(
        kw_only=True,
        default=Context.UNDEFINED,
    )



from xformula.syntax.ast.nodes.abc.has_context import HasContext
from xformula.syntax.ast.nodes.abc.has_name import HasName
from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.term import Term

__all__ = [
    "Identifier",
]


class Identifier(
    Term,
    HasName[str],
    HasContext,
    Node,
):
    class Meta(
        Term.Meta,
        HasName.Meta,
        HasContext.Meta,
        Node.Meta,
    ):

        abstract = True

        is_identifier = True



from xformula.syntax.core.features.expressions.definitions.non_terminals.parenthesized_simple_expression import (
    ParenthesizedSimpleExpression,
)
from xformula.syntax.core.features.expressions.definitions.non_terminals.simple_expression import (
    SimpleExpression,
)

__all__ = [
    "ParenthesizedSimpleExpression",
    "SimpleExpression",
]



from typing import Iterable, cast

__all__ = [
    "suite",
]


def suite(*lines: str) -> str:
    return "\n".join(
        cast(
            Iterable[str],
            filter(
                len,
                lines,
            ),
        ),
    )



import dataclasses
from functools import partial
from itertools import chain
from typing import TYPE_CHECKING, Optional, TypeVar, cast

from xformula.arch.bisect import Bisect

if TYPE_CHECKING:
    from xformula.runtime.core.context.abc.runtime_context import RuntimeContext
    from xformula.syntax.core.features.abc.feature import Feature
    from xformula.syntax.core.operations.abc.operator_precedence import (
        OperatorPrecedence,
    )
    from xformula.syntax.grammar.definitions.abc.definition import Definition
    from xformula.syntax.grammar.definitions.abc.definition_type import DefinitionType
    from xformula.syntax.grammar.definitions.abc.ebnf_expression_builder_protocol import (
        EBNFExpressionBuilderProtocol,
    )
    from xformula.syntax.grammar.definitions.abc.regex_expression_builder_protocol import (
        RegexExpressionBuilderProtocol,
    )
    from xformula.syntax.grammar.directives.abc.directive import Directive
    from xformula.syntax.grammar.non_terminals.abc.non_terminal import NonTerminal
    from xformula.syntax.grammar.templates.abc.template import Template
    from xformula.syntax.grammar.terminals.abc.terminal import Terminal
    from xformula.syntax.parser.transformers.abc.ast_builder_protocol import (
        ASTBuilderProtocol,
    )

__all__ = [
    "SyntaxContext",
]


C = TypeVar("C", bound="SyntaxContext")


@dataclasses.dataclass()
class SyntaxContext:
    @staticmethod
    def build_ebnf_expression_builder_class() -> type[
        "EBNFExpressionBuilderProtocol",
    ]:
        from xformula.syntax.grammar.definitions.runtime.reflection import (
            EBNFExpressionBuilderClassBuilder,
        )

        return EBNFExpressionBuilderClassBuilder.build()

    @staticmethod
    def build_regex_expression_builder_class() -> type[
        "RegexExpressionBuilderProtocol",
    ]:
        from xformula.syntax.grammar.definitions.runtime.reflection import (
            RegexExpressionBuilderClassBuilder,
        )

        return RegexExpressionBuilderClassBuilder.build()

    feature_types: dataclasses.InitVar[
        list[type["Feature"]] | None
    ] = dataclasses.field(
        kw_only=True,
        default=None,
    )

    ebnf_expression_builder_class: type[
        "EBNFExpressionBuilderProtocol",
    ] = dataclasses.field(
        default_factory=build_ebnf_expression_builder_class,
    )

    regex_expression_builder_class: type[
        "RegexExpressionBuilderProtocol",
    ] = dataclasses.field(
        default_factory=build_regex_expression_builder_class,
    )

    features: dict[str, "Feature"] = dataclasses.field(
        kw_only=True,
        default_factory=dict,
    )

    directives: list["Directive"] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )

    non_terminals: list["NonTerminal"] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )

    templates: list["Template"] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )

    terminals: list["Terminal"] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )

    tagged_definitions: dict[str, list["Definition"]] = dataclasses.field(
        kw_only=True,
        default_factory=dict,
    )

    operator_precedences: list["OperatorPrecedence"] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )

    def __post_init__(
        self,
        feature_types: list[type["Feature"]] | None,
    ) -> None:
        if feature_types is not None:
            for feature_type in feature_types:
                self.register_feature_type(feature_type)
        self.setup()

    def get_feature(
        self,
        fqn: str,
    ) -> Optional["Feature"]:
        return self.features.get(fqn)

    def f(
        self,
        fqn: str,
    ) -> "Feature":
        return self.features[fqn]

    def register_feature_type(
        self,
        feature_type: type["Feature"],
    ) -> None:
        feature = feature_type(self)
        self.register_feature(feature)

    def register_feature(
        self,
        feature: "Feature",
    ) -> None:
        self.features[feature.__class__.options.fqn] = feature

    def register_feature_definitions(
        self,
        feature: "Feature",
    ) -> None:
        definition_types = chain(
            feature.directive_types,
            feature.non_terminal_types,
            feature.template_types,
            feature.terminal_types,
        )
        for definition_type in definition_types:
            definition = definition_type(self)
            definitions = self.get_definitions_by_type(definition_type.options.type)
            definitions.append(definition)
            for tag, priority in definition_type.options.tags.items():
                if tag not in self.tagged_definitions:
                    self.tagged_definitions[tag] = []
                bisect_key = partial(self.get_definition_tag_priority, tag)
                Bisect.insert_right(
                    self.tagged_definitions[tag],
                    definition,
                    lhs_key=bisect_key,
                    rhs_key=bisect_key,
                )

    @staticmethod
    def get_definition_tag_priority(
        tag: str,
        definition: "Definition",
    ) -> int | float:
        return definition.__class__.options.tags.get(tag, 0)

    def get_definitions_by_type(
        self,
        definition_type: "DefinitionType",
    ) -> list["Definition"]:
        from xformula.syntax.grammar import DefinitionType

        if definition_type == DefinitionType.DIRECTIVE:
            return cast(list["Definition"], self.directives)
        if definition_type == DefinitionType.NON_TERMINAL:
            return cast(list["Definition"], self.non_terminals)
        if definition_type == DefinitionType.TEMPLATE:
            return cast(list["Definition"], self.templates)
        if definition_type == DefinitionType.TERMINAL:
            return cast(list["Definition"], self.terminals)

        raise TypeError(f"invalid definition type: {definition_type!r}")

    def setup_features(self) -> None:
        # todo: These get done before runtime, we still need a signal-handling mechanism.
        for feature in self.features.values():
            feature.pre_setup()
        for feature in self.features.values():
            feature.setup()
        for feature in self.features.values():
            self.register_feature_definitions(feature)
        for feature in self.features.values():
            feature.post_setup()
        for feature in self.features.values():
            feature.on_ready()

    def setup(self) -> None:
        self.setup_features()

    def build_ast_builder_class(
        self,
        runtime_context: "RuntimeContext",
    ) -> type["ASTBuilderProtocol"]:
        from xformula.syntax.parser.transformers.runtime.reflection import (
            ASTBuilderClassBuilder,
        )

        return ASTBuilderClassBuilder.build(
            self,
            runtime_context,
        )



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "LiteralFeature",
]


class LiteralFeature(Feature):
    class Meta:
        fqn = "xformula.core.Literal"

    def setup(self) -> None:
        from xformula.syntax.core.features.literals.definitions.non_terminals import (
            Bool,
            Complex,
            Float,
            Int,
            Literal,
            None_,
            Str,
        )
        from xformula.syntax.core.features.literals.definitions.terminals import (
            BINARY_INT,
            BOOL,
            COMPLEX,
            DECIMAL_INT,
            FLOAT,
            HEXADECIMAL_INT,
            NONE,
            OCTAL_INT,
            SINGLE_LINE_STR,
        )

        self.non_terminal_types.extend(
            [
                None_,
                Bool,
                Int,
                Float,
                Complex,
                Str,
                Literal,
            ],
        )

        self.terminal_types.extend(
            [
                NONE,
                BOOL,
                BINARY_INT,
                OCTAL_INT,
                DECIMAL_INT,
                HEXADECIMAL_INT,
                FLOAT,
                COMPLEX,
                SINGLE_LINE_STR,
            ],
        )



from xformula.syntax.core.features.abc import Feature, FeatureMeta, FeatureOptions
from xformula.syntax.core.features.attributes import AttributeFeature
from xformula.syntax.core.features.calls import CallFeature
from xformula.syntax.core.features.containers import ContainerFeature
from xformula.syntax.core.features.default_feature_types import DEFAULT_FEATURE_TYPES
from xformula.syntax.core.features.expressions import ExpressionFeature
from xformula.syntax.core.features.identifiers import IdentifierFeature
from xformula.syntax.core.features.literals import LiteralFeature
from xformula.syntax.core.features.mappings import MappingFeature
from xformula.syntax.core.features.operations import OperationFeature
from xformula.syntax.core.features.polyfill import PolyfillFeature
from xformula.syntax.core.features.primaries import PrimaryFeature
from xformula.syntax.core.features.separators import SeparatorFeature
from xformula.syntax.core.features.terms import TermFeature
from xformula.syntax.core.features.validation import ValidationFeature
from xformula.syntax.core.features.wrappers import WrapperFeature

__all__ = [
    "AttributeFeature",
    "CallFeature",
    "ContainerFeature",
    "DEFAULT_FEATURE_TYPES",
    "ExpressionFeature",
    "Feature",
    "FeatureMeta",
    "FeatureOptions",
    "IdentifierFeature",
    "LiteralFeature",
    "MappingFeature",
    "OperationFeature",
    "PolyfillFeature",
    "PrimaryFeature",
    "SeparatorFeature",
    "TermFeature",
    "ValidationFeature",
    "WrapperFeature",
]



from xformula.syntax.ast.nodes.abc.arity import Arity
from xformula.syntax.ast.nodes.abc.node import Node

__all__ = [
    "HasArity",
]


class HasArity(Node):

    arity: Arity

    class Meta(Node.Meta):

        abstract: bool = True

        has_arity_attribute: bool = True



import dataclasses

from xformula.syntax.ast.nodes import Literal

__all__ = [
    "Float",
]


@dataclasses.dataclass()
class Float(
    Literal[float],
):

    value: float = dataclasses.field(
        kw_only=True,
        default=float(),
    )



from xformula.syntax.grammar.templates.abc import Template

__all__ = [
    "WrapZeroOrMoreOf",
]


class WrapZeroOrMoreOf(Template):
    class Meta:

        arguments = [
            Template.arg("left"),
            Template.arg("right"),
            Template.arg("rule"),
        ]

    def build_grammar(self) -> str:
        suite = self.ebnf.suite
        define = self.ebnf.define
        orelse = self.ebnf.orelse
        arg = self.ebnf.arg
        use_template = self.ebnf.use_template

        return suite(
            define(
                use_template(
                    "WrapOneOrMoreOf",
                    arg("left"),
                    arg("right"),
                    arg("rule"),
                ),
            ),
            orelse(arg("left"), arg("right")),
        )



from xformula.syntax.grammar.regex._flat_join import flat_join

__all__ = [
    "chars_between",
]


def chars_between(
    lower: str,
    upper: str,
) -> str:
    return flat_join("-", [lower, upper])



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "ContainerFeature",
]


class ContainerFeature(Feature):
    class Meta:
        fqn = "xformula.core.Container"

    def setup(self) -> None:
        from xformula.syntax.core.features.containers.definitions.non_terminals import (
            List,
            Set,
            Tuple,
        )

        self.non_terminal_types.extend(
            [
                List,
                Tuple,
                Set,
            ],
        )



from abc import ABC

from xformula.syntax.grammar.definitions.abc import Definition, DefinitionType

__all__ = [
    "Directive",
]


class Directive(
    Definition,
    ABC,
):
    class Meta(Definition.Meta):

        abstract = True

        type = DefinitionType.DIRECTIVE



from typing import Generic, TypeVar

from xformula.syntax.ast.nodes.abc.node import Node

__all__ = [
    "HasValue",
]


T = TypeVar("T")


class HasValue(
    Generic[T],
    Node,
):

    value: T

    class Meta(Node.Meta):

        abstract = True

        has_value_attribute = True



from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.core.features.literals.ast.nodes import None_ as NoneNode
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "None_",
]


class None_(  # NOSONAR
    NonTerminal[NoneNode],
):
    class Meta:

        definition_name = "None"

        atomic = True

        tags = {
            non_terminal("Literal"): -1000,
        }

    def build_grammar(self) -> str:
        return self.ebnf.define_tagged_alternation()

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree,
    ) -> NoneNode:
        return NoneNode()



from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.operator import Operator

__all__ = [
    "HasOperator",
]


class HasOperator(Node):

    operator: Operator

    class Meta(Node.Meta):

        abstract = True

        has_operator_attribute = True



from xformula.syntax.grammar.regex._flat_join import flat_join

__all__ = [
    "non_capturing_group",
]


def non_capturing_group(*expressions: str) -> str:
    inner = flat_join("", expressions)
    if not inner:
        return ""
    return f"(?:{inner})"



from typing import Iterable, cast

__all__ = [
    "flat_join",
]


def flat_join(
    separator: str,
    expressions: Iterable[str],
) -> str:
    return separator.join(
        cast(
            Iterable[str],
            filter(
                len,
                expressions,
            ),
        ),
    )



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import Context, Identifier, SimpleExpression
from xformula.syntax.core.features.calls.ast.nodes import (
    CallKeywordArgument as FunctionCallKeywordArgumentNode,
)
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "CallKeywordArgument",
]


class CallKeywordArgument(
    NonTerminal[FunctionCallKeywordArgumentNode],
):
    class Meta:

        atomic = True

    def build_grammar(self) -> str:
        define = self.ebnf.define
        use_template = self.ebnf.use_template
        op = self.ebnf.op
        _non_terminal = self.ebnf.non_terminal

        return define(
            use_template(
                "SeparatedPair",
                op("="),
                _non_terminal("Identifier"),
                _non_terminal("SimpleExpression"),
            ),
        )

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[
            tuple[SimpleExpression, SimpleExpression],
        ],
    ) -> FunctionCallKeywordArgumentNode:
        elements = cast(
            tuple[SimpleExpression, SimpleExpression],
            tuple(tree.children),
        )
        key = elements[0]
        value = elements[1]
        if isinstance(key, Identifier):
            key.context = Context.KEYWORD_ARGUMENT
        return FunctionCallKeywordArgumentNode(
            elements=(key, value),
        )



import dataclasses

__all__ = [
    "Options",
]


@dataclasses.dataclass()
class Options:

    name: str = dataclasses.field(
        kw_only=True,
        default=str(),
    )

    abstract: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )



from xformula.syntax.parser.transformers.abc.ast_builder_protocol import (
    ASTBuilderProtocol,
)
from xformula.syntax.parser.transformers.abc.transformer import Transformer

__all__ = [
    "ASTBuilderProtocol",
    "Transformer",
]



import dataclasses
from typing import Any, Callable

from xformula.arch.meta import Options
from xformula.syntax.grammar.definitions.abc.definition_type import DefinitionType

__all__ = [
    "DefinitionOptions",
]


@dataclasses.dataclass()
class DefinitionOptions(Options):

    definition_name: str = dataclasses.field(
        kw_only=True,
        default=str(),
    )

    atomic: bool = dataclasses.field(
        kw_only=True,
        default=False,
    )

    retain_anonymous_literals: bool = dataclasses.field(
        kw_only=True,
        default=False,
    )

    type: DefinitionType = dataclasses.field(
        kw_only=True,
        default=DefinitionType.UNDEFINED,
    )

    arguments: list[str] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )

    priority: int = dataclasses.field(
        kw_only=True,
        default=int(),
    )

    tags: dict[str, int | float] = dataclasses.field(
        kw_only=True,
        default_factory=dict,
    )

    transform: str | Callable[..., Any] = dataclasses.field(
        kw_only=True,
        default="transform",
    )



from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.grammar.terminals.abc import Terminal
from xformula.syntax.lexer.tokens.abc import Token

__all__ = [
    "WHITESPACE",
]


class WHITESPACE(
    Terminal[str],
):
    def build_grammar(self) -> str:
        define = self.ebnf.define
        regex = self.ebnf.regex

        one_or_more_of = self.regex.one_or_more_of
        whitespace = self.regex.whitespace

        return define(
            regex(
                one_or_more_of(
                    whitespace(),
                ),
            ),
        )

    def transform_token(
        self,
        runtime_context: RuntimeContext,
        token: Token,
    ) -> str:
        return token.value



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "ExpressionFeature",
]


class ExpressionFeature(Feature):
    class Meta:
        fqn = "xformula.core.Expression"

    def setup(self) -> None:
        from xformula.syntax.core.features.expressions.definitions.non_terminals import (
            ParenthesizedSimpleExpression,
            SimpleExpression,
        )

        self.non_terminal_types.extend(
            [
                SimpleExpression,
                ParenthesizedSimpleExpression,
            ],
        )



from xformula.arch.meta.configurable import Configurable
from xformula.arch.meta.meta import Meta
from xformula.arch.meta.options import Options

__all__ = [
    "Configurable",
    "Meta",
    "Options",
]



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "WrapperFeature",
]


class WrapperFeature(Feature):
    class Meta:
        fqn = "xformula.core.Wrapper"

    def setup(self) -> None:
        from xformula.syntax.core.features.wrappers.definitions.templates import (
            WrapOneOrMoreOf,
            WrapZeroOrMoreOf,
        )

        self.template_types.extend(
            [
                WrapZeroOrMoreOf,
                WrapOneOrMoreOf,
            ],
        )



from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.terminals.abc import Terminal
from xformula.syntax.lexer.tokens.abc import Token

__all__ = [
    "FLOAT",
]


class FLOAT(
    Terminal[float],
):
    class Meta:

        priority = 5000

        tags = {
            non_terminal("Float"): 1000,
        }

    def build_grammar(self) -> str:
        define = self.ebnf.define
        regex = self.ebnf.regex

        any_char_of = self.regex.any_char_of
        any_of = self.regex.any_of
        chars_between = self.regex.chars_between
        escape_char = self.regex.escape_char
        flat = self.regex.flat
        flat_optional = self.regex.flat_optional
        flat_zero_or_more_of = self.regex.flat_zero_or_more_of
        non_capturing_group = self.regex.non_capturing_group

        re_number_dot_optional_scale = flat(
            any_char_of(chars_between("0", "9")),
            flat_zero_or_more_of(
                non_capturing_group(
                    flat_optional("_"),
                    any_char_of(chars_between("0", "9")),
                ),
            ),
            escape_char("."),
            flat_optional(
                non_capturing_group(
                    any_char_of(chars_between("0", "9")),
                    flat_zero_or_more_of(
                        non_capturing_group(
                            flat_optional("_"),
                            any_char_of(chars_between("0", "9")),
                        ),
                    ),
                ),
            ),
        )

        re_dot_number = flat(
            escape_char("."),
            any_char_of(chars_between("0", "9")),
            flat_zero_or_more_of(
                non_capturing_group(
                    flat_optional(r"_"),
                    any_char_of(chars_between("0", "9")),
                ),
            ),
        )

        re_exponent_part = flat(
            any_char_of("e", "E"),
            flat_optional(any_char_of("-", "+")),
            any_char_of(chars_between("0", "9")),
            flat_zero_or_more_of(
                non_capturing_group(
                    flat_optional("_"),
                    any_char_of(chars_between("0", "9")),
                ),
            ),
        )

        re_dot_float = flat(
            non_capturing_group(
                any_of(
                    re_number_dot_optional_scale,
                    re_dot_number,
                ),
            ),
            flat_optional(
                non_capturing_group(re_exponent_part),
            ),
        )

        re_number_float = flat(
            any_char_of(chars_between("0", "9")),
            flat_zero_or_more_of(
                non_capturing_group(
                    flat_optional("_"),
                    any_char_of(chars_between("0", "9")),
                ),
            ),
            re_exponent_part,
        )

        return define(regex(any_of(re_dot_float, re_number_float)))

    def transform_token(
        self,
        runtime_context: RuntimeContext,
        token: Token,
    ) -> float:
        return float(token.value)



from xformula.syntax.core.features.literals.ast.nodes.bool import Bool
from xformula.syntax.core.features.literals.ast.nodes.complex import Complex
from xformula.syntax.core.features.literals.ast.nodes.float import Float
from xformula.syntax.core.features.literals.ast.nodes.int import Int
from xformula.syntax.core.features.literals.ast.nodes.none_ import None_
from xformula.syntax.core.features.literals.ast.nodes.str import Str

__all__ = [
    "Bool",
    "Complex",
    "Float",
    "Int",
    "None_",
    "Str",
]



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import Context, HasContext, Identifier, Node
from xformula.syntax.core.features.attributes.ast.nodes import (
    Attribute as AttributeNode,
)
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "Attribute",
]


class Attribute(
    NonTerminal[AttributeNode],
):
    class Meta:

        atomic = True

        tags = {
            non_terminal("Term"): 15_000,
        }

    def build_grammar(self) -> str:
        define = self.ebnf.define
        _non_terminal = self.ebnf.non_terminal
        op = self.ebnf.op
        orelse = self.ebnf.orelse
        suite = self.ebnf.suite

        return suite(
            define(_non_terminal("Attribute"), op("."), _non_terminal("Identifier")),
            orelse(_non_terminal("Primary"), op("."), _non_terminal("Identifier")),
        )

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[Node],
    ) -> AttributeNode:
        owner = cast(Node, tree.children[0])
        if owner.__class__.options.has_context_attribute:
            owner = cast(HasContext, owner)
            owner.context = getattr(
                owner.context.__class__,
                "LOAD",
                Context.LOAD,
            )
        name = cast(Identifier, tree.children[1])
        name.context = name.context.__class__.ATTRIBUTE
        return AttributeNode(
            owner=owner,
            name=name,
            context=Context.LOAD,
        )



from xformula.syntax.core.features.literals.definitions.non_terminals.bool import Bool
from xformula.syntax.core.features.literals.definitions.non_terminals.complex import (
    Complex,
)
from xformula.syntax.core.features.literals.definitions.non_terminals.float import Float
from xformula.syntax.core.features.literals.definitions.non_terminals.int import Int
from xformula.syntax.core.features.literals.definitions.non_terminals.literal import (
    Literal,
)
from xformula.syntax.core.features.literals.definitions.non_terminals.none_ import None_
from xformula.syntax.core.features.literals.definitions.non_terminals.str import Str

__all__ = [
    "Bool",
    "Complex",
    "Float",
    "Int",
    "Literal",
    "None_",
    "Str",
]



from xformula.syntax.core.features.calls.ast.nodes.call import Call
from xformula.syntax.core.features.calls.ast.nodes.call_keyword_argument import (
    CallKeywordArgument,
)

__all__ = [
    "Call",
    "CallKeywordArgument",
]



from typing import TypeVar, cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import Literal as LiteralNode
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "Literal",
]


T = TypeVar("T")


class Literal(
    NonTerminal[
        LiteralNode[T],
    ],
):
    class Meta:

        tags = {
            non_terminal("Term"): 5_000,
        }

    def build_grammar(self) -> str:
        return self.ebnf.define_tagged_alternation()

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[T],
    ) -> LiteralNode[T]:
        return cast(LiteralNode, tree.children[0])



from itertools import chain
from typing import Any

from xformula.arch.runtime.reflection import ClassBuilder
from xformula.syntax.grammar.non_terminals.abc import NonTerminal

__all__ = [
    "NonTerminalClassBuilder",
]


class NonTerminalClassBuilder(ClassBuilder):
    @classmethod
    def get_name(
        cls,
        name: str,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        return name

    @classmethod
    def get_bases(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[type, ...]:
        return (NonTerminal,)

    @classmethod
    def get_namespace(
        cls,
        name: str,
        alternation: set[str],
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Any]:

        branches = list(alternation)

        meta = type(
            "Meta",
            (),
            dict(
                atomic=False,
                is_polyfill=True,
                definition_name=name,
            ),
        )

        def build_grammar(self: NonTerminal) -> str:
            nonlocal branches

            suite = self.ebnf.suite
            define = self.ebnf.define
            orelse = self.ebnf.orelse

            return suite(
                *chain(
                    [
                        define(branches[0]),
                    ],
                    [orelse(branch) for branch in branches[1:]],
                ),
            )

        namespace = dict(
            Meta=meta,
            build_grammar=build_grammar,
        )

        return namespace



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import Context
from xformula.syntax.core.features.identifiers.ast.nodes import (
    Identifier as IdentifierNode,
)
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "Identifier",
]


class Identifier(
    NonTerminal[IdentifierNode],
):
    class Meta:

        atomic = True

        tags = {
            non_terminal("Term"): 10_000,
        }

    def build_grammar(self) -> str:
        return self.ebnf.define_tagged_alternation()

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[str],
    ) -> IdentifierNode:
        name = cast(str, tree.children[0])
        return IdentifierNode(
            name=name,
            context=Context.LOAD,
        )



from typing import TYPE_CHECKING, Iterable, Protocol, overload, runtime_checkable

if TYPE_CHECKING:
    from xformula.syntax.grammar.definitions.abc.definition import Definition

__all__ = [
    "RegexExpressionBuilderProtocol",
]


@runtime_checkable
class RegexExpressionBuilderProtocol(Protocol):

    _definition: "Definition"

    def __init__(
        self,
        definition: "Definition",
    ) -> None:
        self._definition = definition

    def any_char(self) -> str:
        ...

    def any_char_except(
        self,
        *chars: str,
    ) -> str:
        ...

    def any_char_of(
        self,
        *chars: str,
    ) -> str:
        ...

    def any_of(
        self,
        *expressions: str,
    ) -> str:
        ...

    def bound(
        self,
        *expressions: str,
    ) -> str:
        ...

    def chars_between(
        self,
        lower: str,
        upper: str,
    ) -> str:
        ...

    def escape_char(
        self,
        char: str,
    ) -> str:
        ...

    def flat(
        self,
        *expressions: str,
    ) -> str:
        ...

    def flat_join(
        self,
        separator: str,
        expressions: Iterable[str],
    ) -> str:
        ...

    def flat_one_or_more_of(
        self,
        expression: str,
    ) -> str:
        ...

    def flat_optional(
        self,
        expression: str,
    ) -> str:
        ...

    def flat_zero_or_more_of(
        self,
        expression: str,
    ) -> str:
        ...

    def non_capturing_group(
        self,
        *expressions: str,
    ) -> str:
        ...

    def one_or_more_of(
        self,
        *expressions: str,
    ) -> str:
        ...

    def whitespace(self) -> str:
        ...

    def word(
        self,
        chars: str,
    ) -> str:
        ...

    @overload
    def wrap(
        self,
        outer: str,
        inner: str,
    ) -> str:
        ...

    @overload
    def wrap(
        self,
        left: str,
        right: str,
        inner: str,
    ) -> str:
        ...

    @overload
    def wrap(
        self,
        left: str,
        right: str,
        *inner: str,
    ) -> str:
        ...

    def wrap(
        self,
        *args,
    ):
        ...

    def zero_or_more_of(
        self,
        *expressions: str,
    ) -> str:
        ...



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import Node
from xformula.syntax.core.features.calls.ast.nodes import Call as FunctionCallNode
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "Call",
]


class Call(
    NonTerminal[FunctionCallNode],
):
    class Meta:

        atomic = True

        tags = {
            non_terminal("Term"): 20_000,
        }

    def build_grammar(self) -> str:
        define = self.ebnf.define
        _non_terminal = self.ebnf.non_terminal
        op = self.ebnf.op
        optional = self.ebnf.optional
        orelse = self.ebnf.orelse
        use_template = self.ebnf.use_template
        suite = self.ebnf.suite

        return suite(
            define(
                _non_terminal("Term"),
                op("("),
                use_template(
                    "SeparatedZeroOrMoreOf",
                    op(","),
                    _non_terminal("CallArgument"),
                ),
                op(","),
                use_template(
                    "SeparatedZeroOrMoreOf",
                    op(","),
                    _non_terminal("CallKeywordArgument"),
                ),
                optional(op(",")),
                op(")"),
            ),
            orelse(
                _non_terminal("Term"),
                op("("),
                use_template(
                    "SeparatedZeroOrMoreOf",
                    op(","),
                    _non_terminal("CallArgument"),
                ),
                optional(op(",")),
                op(")"),
            ),
            orelse(
                _non_terminal("Term"),
                op("("),
                use_template(
                    "SeparatedZeroOrMoreOf",
                    op(","),
                    _non_terminal("CallKeywordArgument"),
                ),
                optional(op(",")),
                op(")"),
            ),
            orelse(_non_terminal("Term"), op("("), op(")")),
        )

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[Node],
    ) -> FunctionCallNode:
        callee = cast(Node, tree.children[0])
        arguments = cast(list[Node], tree.children[1:])
        return FunctionCallNode(
            callee=callee,
            arguments=arguments,
        )



from typing import Any, Callable, Concatenate, ParamSpec, TypeVar

from xformula.arch.runtime.reflection import ClassBuilder
from xformula.syntax.grammar.definitions.abc import RegexExpressionBuilderProtocol

__all__ = [
    "RegexExpressionBuilderClassBuilder",
]


P = ParamSpec("P")

R = TypeVar("R")


class RegexExpressionBuilderClassBuilder(
    ClassBuilder[RegexExpressionBuilderProtocol],
):
    @classmethod
    def get_name(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        return "RegexExpressionBuilder"

    @classmethod
    def get_bases(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[type, ...]:
        return (RegexExpressionBuilderProtocol,)

    @classmethod
    def get_namespace(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Any]:
        from xformula.syntax.grammar import regex

        namespace: dict[str, Any] = dict()

        for attname in dir(regex):
            if attname.startswith("_"):
                continue
            value = getattr(regex, attname)
            if not callable(value):
                continue
            delegate = cls.build_regex_function_delegator(
                attname,
                value,
            )
            namespace[attname] = delegate

        return namespace

    @classmethod
    def build_regex_function_delegator(
        cls,
        regex_function_name: str,
        regex_function: Callable[P, R],
    ) -> Callable[Concatenate[RegexExpressionBuilderProtocol, P], R]:
        def delegate(
            protocol: RegexExpressionBuilderProtocol,
            *args: P.args,
            **kwargs: P.kwargs,
        ) -> R:
            pre_handler_function_name = f"pre_call_regex_{regex_function_name}"
            pre_handler_function = getattr(protocol, pre_handler_function_name, None)
            post_handler_function_name = f"post_call_regex_{regex_function_name}"
            post_handler_function = getattr(protocol, post_handler_function_name, None)
            if callable(pre_handler_function):
                args, kwargs = pre_handler_function(args, kwargs)
            result = regex_function(*args, **kwargs)
            if callable(post_handler_function):
                post_handler_function(args, kwargs, result)
            return result

        if hasattr(regex_function, "__name__"):
            delegate.__name__ = getattr(regex_function, "__name__")

        return delegate



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "SeparatorFeature",
]


class SeparatorFeature(Feature):
    class Meta:
        fqn = "xformula.core.Separator"

    def setup(self) -> None:
        from xformula.syntax.core.features.separators.definitions.templates import (
            SeparatedOneOrMoreOf,
            SeparatedPair,
            SeparatedZeroOrMoreOf,
        )

        self.template_types.extend(
            [
                SeparatedPair,
                SeparatedZeroOrMoreOf,
                SeparatedOneOrMoreOf,
            ],
        )



from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.primary import Primary

__all__ = [
    "Term",
]


class Term(
    Primary,
    Node,
):
    class Meta(
        Primary.Meta,
        Node.Meta,
    ):

        abstract = True

        is_term = True



from xformula.syntax.grammar.definitions.abc.definition import Definition
from xformula.syntax.grammar.definitions.abc.definition_meta import DefinitionMeta
from xformula.syntax.grammar.definitions.abc.definition_options import DefinitionOptions
from xformula.syntax.grammar.definitions.abc.definition_type import DefinitionType
from xformula.syntax.grammar.definitions.abc.ebnf_expression_builder_protocol import (
    EBNFExpressionBuilderProtocol,
)
from xformula.syntax.grammar.definitions.abc.regex_expression_builder_protocol import (
    RegexExpressionBuilderProtocol,
)

__all__ = [
    "Definition",
    "DefinitionMeta",
    "DefinitionOptions",
    "DefinitionType",
    "EBNFExpressionBuilderProtocol",
    "RegexExpressionBuilderProtocol",
]



from typing import Generic, Sequence, TypeVar

from xformula.syntax.ast.nodes.abc.node import Node

__all__ = [
    "HasElements",
]


S = TypeVar("S", bound=Sequence)


class HasElements(
    Generic[S],
    Node,
):

    elements: S

    class Meta(Node.Meta):

        abstract = True

        has_elements_attribute = True



from abc import ABC
from typing import Generic, TypeVar

from xformula.syntax.grammar.definitions.abc import Definition, DefinitionType

__all__ = [
    "Template",
]


T = TypeVar("T")


class Template(
    Definition,
    ABC,
    Generic[T],
):
    class Meta(Definition.Meta):

        abstract = True

        type = DefinitionType.TEMPLATE

    @staticmethod
    def arg(name: str) -> str:
        from xformula.syntax.grammar.ebnf import arg

        return arg(name)



from abc import ABC
from typing import Generic, TypeVar

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.grammar.definitions.abc import Definition, DefinitionType
from xformula.syntax.lexer.tokens.abc import Token

__all__ = [
    "Terminal",
]


T = TypeVar("T")


class Terminal(
    Definition,
    Generic[T],
    ABC,
):
    class Meta(Definition.Meta):

        abstract = True

        atomic = True

        type = DefinitionType.TERMINAL

        transform = "transform_token"

    def transform_token(
        self,
        runtime_context: RuntimeContext,
        token: Token,
    ) -> T:
        raise NotImplementedError(
            f"{self.__class__.__qualname__}.transform_token is not implemented",
        )



from xformula.syntax.ast.nodes.abc.arity import Arity
from xformula.syntax.ast.nodes.abc.associativity import Associativity
from xformula.syntax.ast.nodes.abc.attribute import Attribute
from xformula.syntax.ast.nodes.abc.call import Call
from xformula.syntax.ast.nodes.abc.container import Container
from xformula.syntax.ast.nodes.abc.context import Context
from xformula.syntax.ast.nodes.abc.dict import Dict
from xformula.syntax.ast.nodes.abc.expression import Expression
from xformula.syntax.ast.nodes.abc.has_arguments import HasArguments
from xformula.syntax.ast.nodes.abc.has_arity import HasArity
from xformula.syntax.ast.nodes.abc.has_associativity import HasAssociativity
from xformula.syntax.ast.nodes.abc.has_callee import HasCallee
from xformula.syntax.ast.nodes.abc.has_context import HasContext
from xformula.syntax.ast.nodes.abc.has_elements import HasElements
from xformula.syntax.ast.nodes.abc.has_name import HasName
from xformula.syntax.ast.nodes.abc.has_operands import HasOperands
from xformula.syntax.ast.nodes.abc.has_operator import HasOperator
from xformula.syntax.ast.nodes.abc.has_owner import HasOwner
from xformula.syntax.ast.nodes.abc.has_placement import HasPlacement
from xformula.syntax.ast.nodes.abc.has_precedence_level import HasPrecedenceLevel
from xformula.syntax.ast.nodes.abc.has_symbols import HasSymbols
from xformula.syntax.ast.nodes.abc.has_value import HasValue
from xformula.syntax.ast.nodes.abc.identifier import Identifier
from xformula.syntax.ast.nodes.abc.list import List
from xformula.syntax.ast.nodes.abc.literal import Literal
from xformula.syntax.ast.nodes.abc.mapping import Mapping
from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.node_meta import NodeMeta
from xformula.syntax.ast.nodes.abc.node_options import NodeOptions
from xformula.syntax.ast.nodes.abc.operand import Operand
from xformula.syntax.ast.nodes.abc.operation import Operation
from xformula.syntax.ast.nodes.abc.operator import Operator
from xformula.syntax.ast.nodes.abc.pair import Pair
from xformula.syntax.ast.nodes.abc.placement import Placement
from xformula.syntax.ast.nodes.abc.precedence_level import PrecedenceLevel
from xformula.syntax.ast.nodes.abc.primary import Primary
from xformula.syntax.ast.nodes.abc.set import Set
from xformula.syntax.ast.nodes.abc.simple_expression import SimpleExpression
from xformula.syntax.ast.nodes.abc.symbol import Symbol
from xformula.syntax.ast.nodes.abc.symbol_type import SymbolType
from xformula.syntax.ast.nodes.abc.term import Term
from xformula.syntax.ast.nodes.abc.tuple import Tuple

__all__ = [
    "Arity",
    "Associativity",
    "Attribute",
    "Call",
    "Container",
    "Context",
    "Dict",
    "Expression",
    "HasArguments",
    "HasArity",
    "HasAssociativity",
    "HasCallee",
    "HasContext",
    "HasElements",
    "HasName",
    "HasOperands",
    "HasOperator",
    "HasOwner",
    "HasPlacement",
    "HasPrecedenceLevel",
    "HasSymbols",
    "HasValue",
    "Identifier",
    "List",
    "Literal",
    "Mapping",
    "Node",
    "NodeMeta",
    "NodeOptions",
    "Operand",
    "Operation",
    "Operator",
    "Pair",
    "Placement",
    "PrecedenceLevel",
    "Primary",
    "Set",
    "SimpleExpression",
    "Symbol",
    "SymbolType",
    "Term",
    "Tuple",
]



from xformula.syntax.ast.nodes.abc.has_arguments import HasArguments
from xformula.syntax.ast.nodes.abc.has_callee import HasCallee
from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.term import Term

__all__ = [
    "Call",
]


class Call(
    Term,
    HasCallee[Node],
    HasArguments[list[Node]],
    Node,
):
    class Meta(
        Term.Meta,
        HasCallee.Meta,
        HasArguments.Meta,
        Node.Meta,
    ):

        abstract = True

        is_call = True



from xformula.syntax.ast.nodes.abc.associativity import Associativity
from xformula.syntax.ast.nodes.abc.node import Node

__all__ = [
    "HasAssociativity",
]


class HasAssociativity(Node):

    associativity: Associativity

    class Meta(Node.Meta):

        abstract = True

        has_associativity_attribute = True



from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.terminals.abc import Terminal
from xformula.syntax.lexer.tokens.abc import Token

__all__ = [
    "BOOL",
]


class BOOL(
    Terminal[bool],
):
    class Meta:

        priority = 2000

        tags = {
            non_terminal("Bool"): 1000,
        }

    def build_grammar(self) -> str:
        define = self.ebnf.define
        regex = self.ebnf.regex

        any_of = self.regex.any_of
        bound = self.regex.bound
        word = self.regex.word

        return define(
            regex(
                any_of(
                    bound(word("false")),
                    bound(word("true")),
                ),
            ),
        )

    def transform_token(
        self,
        runtime_context: RuntimeContext,
        token: Token,
    ) -> bool:
        return token.value.lower() == "true"



import dataclasses

from xformula.arch.meta import Options

__all__ = [
    "FeatureOptions",
]


@dataclasses.dataclass()
class FeatureOptions(Options):

    fqn: str = dataclasses.field(
        kw_only=True,
        default=str(),
    )



from itertools import chain
from typing import TYPE_CHECKING

from xformula.syntax.grammar.ebnf._define_rule import define_rule

if TYPE_CHECKING:
    from xformula.syntax.core.context.abc.syntax_context import SyntaxContext

__all__ = [
    "document",
]


def document(syntax_context: "SyntaxContext") -> str:
    return "\n".join(
        chain(
            [
                "// Auto-generated via xformula.syntax module",
            ],
            [""],
            [
                f"{directive.build_grammar()}\n"
                for directive in syntax_context.directives[::-1]
            ],
            [""],
            [
                f"{define_rule(non_terminal)}\n"
                for non_terminal in syntax_context.non_terminals[::-1]
            ],
            [""],
            [
                f"{define_rule(template)}\n"
                for template in syntax_context.templates[::-1]
            ],
            [""],
            [
                f"{define_rule(terminal)}\n"
                for terminal in syntax_context.terminals[::-1]
            ],
        ),
    )



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "AttributeFeature",
]


class AttributeFeature(Feature):
    class Meta:
        fqn = "xformula.core.Attribute"

    def setup(self) -> None:
        from xformula.syntax.core.features.attributes.definitions.non_terminals import (
            Attribute,
        )

        self.non_terminal_types.append(Attribute)



import dataclasses

from xformula.syntax.ast.nodes import Literal

__all__ = [
    "Bool",
]


@dataclasses.dataclass()
class Bool(
    Literal[bool],
):

    value: bool = dataclasses.field(
        kw_only=True,
        default=bool(),
    )



import dataclasses

from xformula.syntax.ast.nodes import Literal

__all__ = [
    "Complex",
]


@dataclasses.dataclass()
class Complex(
    Literal[complex],
):

    value: complex = dataclasses.field(
        kw_only=True,
        default=complex(),
    )



from enum import IntEnum

from xformula.arch.enumeration import Iota

__all__ = [
    "DefinitionType",
]


__xf__definition_type__iota__ = Iota()


class DefinitionType(IntEnum):

    UNDEFINED = __xf__definition_type__iota__.value

    __XF__START__ = __xf__definition_type__iota__()

    DIRECTIVE = __xf__definition_type__iota__()

    NON_TERMINAL = __xf__definition_type__iota__()

    TEMPLATE = __xf__definition_type__iota__()

    TERMINAL = __xf__definition_type__iota__()

    __XF__END__ = __xf__definition_type__iota__()


del __xf__definition_type__iota__



from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.placement import Placement

__all__ = [
    "HasPlacement",
]


class HasPlacement(Node):

    placement: Placement

    class Meta(Node.Meta):

        abstract = True

        has_placement_attribute = True



from typing import NoReturn

from xformula.syntax.core.features.abc import Feature

__all__ = [
    "ValidationFeature",
]

from xformula.syntax.core.features.validation.errors import MissingDefinitionError


class ValidationFeature(Feature):
    class Meta:
        fqn = "xformula.core.Validation"

    def on_ready(self) -> None:
        self.validate()

    def validate(self) -> NoReturn | None:
        from xformula.syntax.core.features.validation.detectors import (
            MissingNonTerminalDetector,
        )

        missing_non_terminals = MissingNonTerminalDetector.detect(
            self.context,
        )

        if len(missing_non_terminals) > 0:
            raise MissingDefinitionError(
                {
                    "non-terminal": missing_non_terminals,
                },
            )

        return None



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "TermFeature",
]


class TermFeature(Feature):
    class Meta:
        fqn = "xformula.core.Term"

    def setup(self) -> None:
        from xformula.syntax.core.features.terms.definitions.non_terminals import Term

        self.non_terminal_types.append(Term)



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "CallFeature",
]


class CallFeature(Feature):
    class Meta:
        fqn = "xformula.core.Call"

    def setup(self) -> None:
        from xformula.syntax.core.features.calls.definitions.non_terminals import (
            Call,
            CallArgument,
            CallKeywordArgument,
        )

        self.non_terminal_types.extend(
            [
                CallArgument,
                CallKeywordArgument,
                Call,
            ],
        )



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import SimpleExpression
from xformula.syntax.core.features.containers.ast.nodes import Tuple as TupleNode
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "Tuple",
]


class Tuple(
    NonTerminal[TupleNode],
):
    class Meta:

        atomic = True

        tags = {
            non_terminal("Primary"): 4000,
        }

    def build_grammar(self) -> str:
        suite = self.ebnf.suite
        define = self.ebnf.define
        orelse = self.ebnf.orelse
        use_template = self.ebnf.use_template
        op = self.ebnf.op
        _non_terminal = self.ebnf.non_terminal
        optional = self.ebnf.optional

        return suite(
            define(
                op("("),
                use_template(
                    "SeparatedOneOrMoreOf", op(","), _non_terminal("SimpleExpression")
                ),
                optional(op(",")),
                op(")"),
            ),
            orelse(op("("), _non_terminal("SimpleExpression"), op(","), op(")")),
            orelse(op("("), op(")")),
        )

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[SimpleExpression],
    ) -> TupleNode:
        return TupleNode(
            elements=cast(list[SimpleExpression], tree.children),
        )



from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.operand import Operand

__all__ = [
    "HasOperands",
]


class HasOperands(Node):

    operands: list[Operand]

    class Meta(Node.Meta):

        abstract = True

        has_operands_attribute = True



from xformula.arch.meta import Meta
from xformula.syntax.ast.nodes.abc.node_options import NodeOptions

__all__ = [
    "NodeMeta",
]


class NodeMeta(
    Meta[NodeOptions],
):

    options_class = NodeOptions

    options: NodeOptions



from xformula.syntax.core.operations.abc import OperatorPrecedence
from xformula.syntax.core.operations.default_operator_precedences import (
    DEFAULT_OPERATOR_PRECEDENCES,
)

__all__ = [
    "DEFAULT_OPERATOR_PRECEDENCES",
    "OperatorPrecedence",
]



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.core.features.literals.ast.nodes import Int as IntNode
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "Int",
]


class Int(
    NonTerminal[IntNode],
):
    class Meta:

        atomic = True

        tags = {
            non_terminal("Literal"): -3000,
        }

    def build_grammar(self) -> str:
        return self.ebnf.define_tagged_alternation()

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[int],
    ) -> IntNode:
        value = cast(int, tree.children[0])
        return IntNode(
            value=value,
        )



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "PolyfillFeature",
]


class PolyfillFeature(Feature):
    class Meta:
        fqn = "xformula.core.Polyfill"

    def post_setup(self) -> None:
        from xformula.syntax.core.features.polyfill.generators import (
            NonTerminalTypeGenerator,
        )
        from xformula.syntax.core.features.validation.detectors import (
            MissingNonTerminalDetector,
        )

        missing_non_terminal_definitions = MissingNonTerminalDetector.detect(
            self.context,
        )

        generated_non_terminal_types = NonTerminalTypeGenerator.generate(
            self.context,
            missing_non_terminal_definitions,
        )

        self.non_terminal_types.extend(generated_non_terminal_types)

        self.context.register_feature_definitions(self)



from xformula.syntax.grammar.definitions.abc import (
    Definition,
    DefinitionMeta,
    DefinitionOptions,
    DefinitionType,
    EBNFExpressionBuilderProtocol,
    RegexExpressionBuilderProtocol,
)

__all__ = [
    "Definition",
    "DefinitionMeta",
    "DefinitionOptions",
    "DefinitionType",
    "EBNFExpressionBuilderProtocol",
    "RegexExpressionBuilderProtocol",
]



from itertools import chain, repeat
from typing import Any, Callable, Iterable, Literal, Union, cast

from xformula.arch.enumeration import Zip
from xformula.arch.runtime.reflection import ClassBuilder
from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import Associativity, Operand, Placement
from xformula.syntax.core.features.operations.ast.nodes import (
    Operation,
    Operator,
    Symbol,
)
from xformula.syntax.core.operations.abc import OperatorPrecedence
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.lexer.tokens.abc import Token
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "NonTerminalOperationClassBuilder",
]


class NonTerminalOperationClassBuilder(
    ClassBuilder[NonTerminal],
):
    @classmethod
    def get_name(
        cls,
        name: str,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        return name

    @classmethod
    def get_bases(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[type, ...]:
        return (NonTerminal,)

    @classmethod
    def get_namespace(
        cls,
        name: str,
        operator_precedence: OperatorPrecedence,
        operand: str,
        *args: Any,
        tags: dict[str, int] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        namespace: dict[str, Any] = dict()

        meta_kwargs = dict(tags=tags) if tags is not None else dict()
        namespace["Meta"] = cls.get_meta_class(operator_precedence, **meta_kwargs)

        namespace["build_grammar"] = cls.get_grammar_builder_function(
            operator_precedence,
            operand,
        )

        namespace["transform_parse_tree"] = cls.get_parse_tree_transformer_function(
            operator_precedence,
            operand,
        )

        return namespace

    @classmethod
    def get_meta_class(
        cls,
        operator_precedence: OperatorPrecedence,
        tags: dict[str, int] | None = None,
    ) -> type:
        if tags is None:
            tags = dict()
        meta_namespace = dict(
            atomic=False,
            priority=operator_precedence.level,
            retain_anonymous_literals=True,
            tags=tags,
        )
        return type("Meta", (), meta_namespace)

    @classmethod
    def cast_zip_join_separator_placement(
        cls,
        placement: Placement,
    ) -> Union[
        Literal[Zip.SeparatorPlacement.LEFT],
        Literal[Zip.SeparatorPlacement.RIGHT],
    ]:
        if placement == Placement.LEFT:
            return Zip.SeparatorPlacement.LEFT
        if placement == Placement.RIGHT:
            return Zip.SeparatorPlacement.RIGHT
        raise TypeError(f"invalid operator placement: {placement!r}")

    @classmethod
    def get_grammar_builder_function(
        cls,
        operator_precedence: OperatorPrecedence,
        operand: str,
    ) -> Callable[[NonTerminal], str]:
        args = (operator_precedence, operand)
        if operator_precedence.arity == 1:
            return cls.get_arity_1_grammar_builder_function(*args)
        return cls.get_arity_n_grammar_builder_function(*args)

    @classmethod
    def get_arity_1_grammar_builder_function(
        cls,
        operator_precedence: OperatorPrecedence,
        operand: str,
    ) -> Callable[[NonTerminal], str]:

        separator_placement = cls.cast_zip_join_separator_placement(
            operator_precedence.placement,
        )

        def build_grammar(self: NonTerminal) -> str:
            recursive_part = self.ebnf.concat(
                chain.from_iterable(
                    Zip.join(
                        [self.ebnf.sym(operator_precedence.symbols[0])],
                        [self.__class__.options.definition_name],
                        separator_placement=separator_placement,
                        default_value="",
                    ),
                ),
            )

            suite = self.ebnf.suite
            define = self.ebnf.define
            orelse = self.ebnf.orelse

            return suite(
                define(recursive_part),
                orelse(operand),
            )

        return build_grammar

    @classmethod
    def get_arity_n_grammar_builder_function(
        cls,
        operator_precedence: OperatorPrecedence,
        operand: str,
    ) -> Callable[[NonTerminal], str]:
        args = (operator_precedence, operand)
        if operator_precedence.associativity == Associativity.LEFT_TO_RIGHT:
            return cls.get_arity_n_ltr_grammar_builder_function(*args)
        if operator_precedence.associativity == Associativity.RIGHT_TO_LEFT:
            return cls.get_arity_n_rtl_grammar_builder_function(*args)
        raise TypeError(
            f"invalid operation associativity: {operator_precedence.associativity!r}"
        )

    @classmethod
    def get_arity_n_ltr_grammar_builder_function(
        cls,
        operator_precedence: OperatorPrecedence,
        operand: str,
    ) -> Callable[[NonTerminal], str]:

        separator_placement = cls.cast_zip_join_separator_placement(
            operator_precedence.placement,
        )

        def build_grammar(self: NonTerminal) -> str:
            left_recursive_part = self.ebnf.concat(
                chain.from_iterable(
                    Zip.join(
                        map(self.ebnf.sym, operator_precedence.symbols),
                        repeat(
                            self.__class__.options.definition_name,
                            times=operator_precedence.arity - 1,
                        ),
                        separator_placement=separator_placement,
                        default_value="",
                    ),
                ),
            )

            define = self.ebnf.define
            optional = self.ebnf.optional

            return define(
                optional(left_recursive_part),
                operand,
            )

        return build_grammar

    @classmethod
    def get_arity_n_rtl_grammar_builder_function(
        cls,
        operator_precedence: OperatorPrecedence,
        operand: str,
    ) -> Callable[[NonTerminal], str]:

        separator_placement = cls.cast_zip_join_separator_placement(
            operator_precedence.placement,
        )

        def build_grammar(self: NonTerminal) -> str:
            left_recursive_part = self.ebnf.concat(
                chain.from_iterable(
                    Zip.join(
                        map(self.ebnf.sym, operator_precedence.symbols),
                        chain(
                            repeat(
                                operand,
                                times=operator_precedence.arity - 2,
                            ),
                            [self.__class__.options.definition_name],
                        ),
                        separator_placement=separator_placement,
                        default_value="",
                    ),
                ),
            )

            define = self.ebnf.define
            optional = self.ebnf.optional

            return define(
                operand,
                optional(left_recursive_part),
            )

        return build_grammar

    @classmethod
    def get_parse_tree_transformer_function(
        cls,
        operator_precedence: OperatorPrecedence,
        operand: str,
    ) -> Callable[[NonTerminal, RuntimeContext, ParseTree], Operation]:
        separator_placement = cls.cast_zip_join_separator_placement(
            operator_precedence.placement,
        )

        if operator_precedence.arity > 1:
            separator_placement = Zip.SeparatorPlacement.RIGHT

        symbol_to_symbol_type_dispatcher = {
            symbol[1]: symbol[0] for symbol in operator_precedence.symbols
        }

        def transform_parse_tree(
            self: NonTerminal,
            runtime_context: RuntimeContext,
            tree: ParseTree,
        ) -> Operation:
            channels = cast(
                tuple[Iterable[Token], Iterable[Operand]],
                Zip.split(
                    cast(Iterable[Token | Operand], tree.children),
                    separator_placement=separator_placement,
                ),
            )
            operator_channel = channels[0]
            operand_channel = channels[1]
            operator = Operator(
                arity=operator_precedence.arity,
                associativity=operator_precedence.associativity,
                placement=operator_precedence.placement,
                precedence_level=operator_precedence.level,
                symbols=[
                    Symbol(
                        type=symbol_to_symbol_type_dispatcher[operator.value],
                        value=operator.value,
                    )
                    for operator in operator_channel
                ],
            )
            return Operation(
                operator=operator,
                operands=cast(list[Operand], list(operand_channel)),
            )

        return transform_parse_tree



from xformula.syntax.core.features.operations.generators.non_terminal_operation_type_generator import (
    NonTerminalOperationTypeGenerator,
)

__all__ = [
    "NonTerminalOperationTypeGenerator",
]



from abc import ABC
from typing import Protocol, runtime_checkable

from xformula.arch.meta.meta import Meta

__all__ = [
    "Configurable",
]


class Configurable(
    ABC,
    metaclass=Meta,
):
    @runtime_checkable
    class Meta(Protocol):

        abstract: bool = True

        name: str = ""



from xformula.syntax.ast.nodes import (
    Arity,
    Associativity,
    Attribute,
    Call,
    Container,
    Context,
    Dict,
    Expression,
    HasArguments,
    HasArity,
    HasAssociativity,
    HasCallee,
    HasContext,
    HasElements,
    HasName,
    HasOperands,
    HasOperator,
    HasOwner,
    HasPlacement,
    HasPrecedenceLevel,
    HasSymbols,
    HasValue,
    Identifier,
    List,
    Literal,
    Mapping,
    Node,
    NodeMeta,
    NodeOptions,
    Operand,
    Operation,
    Operator,
    Pair,
    Placement,
    PrecedenceLevel,
    Primary,
    Set,
    SimpleExpression,
    Symbol,
    SymbolType,
    Term,
    Tuple,
)
from xformula.syntax.core import (
    DEFAULT_FEATURE_TYPES,
    DEFAULT_OPERATOR_PRECEDENCES,
    Feature,
    FeatureMeta,
    FeatureOptions,
    OperatorPrecedence,
    SyntaxContext,
    TaggedDefinitionIterator,
)
from xformula.syntax.grammar import (
    Definition,
    DefinitionMeta,
    DefinitionOptions,
    DefinitionType,
    Directive,
    EBNFExpressionBuilderProtocol,
    NonTerminal,
    RegexExpressionBuilderProtocol,
    Template,
    Terminal,
)
from xformula.syntax.lexer import Token
from xformula.syntax.parser import ASTBuilderProtocol, Parser, ParseTree, Transformer

__all__ = [
    "ASTBuilderProtocol",
    "Arity",
    "Associativity",
    "Attribute",
    "Call",
    "Container",
    "Context",
    "DEFAULT_FEATURE_TYPES",
    "DEFAULT_OPERATOR_PRECEDENCES",
    "Definition",
    "DefinitionMeta",
    "DefinitionOptions",
    "DefinitionType",
    "Dict",
    "Directive",
    "EBNFExpressionBuilderProtocol",
    "Expression",
    "Feature",
    "FeatureMeta",
    "FeatureOptions",
    "HasArguments",
    "HasArity",
    "HasAssociativity",
    "HasCallee",
    "HasContext",
    "HasElements",
    "HasName",
    "HasOperands",
    "HasOperator",
    "HasOwner",
    "HasPlacement",
    "HasPrecedenceLevel",
    "HasSymbols",
    "HasValue",
    "Identifier",
    "List",
    "Literal",
    "Mapping",
    "Node",
    "NodeMeta",
    "NodeOptions",
    "NonTerminal",
    "Operand",
    "Operation",
    "Operator",
    "OperatorPrecedence",
    "Pair",
    "ParseTree",
    "Parser",
    "Placement",
    "PrecedenceLevel",
    "Primary",
    "RegexExpressionBuilderProtocol",
    "Set",
    "SimpleExpression",
    "Symbol",
    "SymbolType",
    "SyntaxContext",
    "TaggedDefinitionIterator",
    "Template",
    "Term",
    "Terminal",
    "Token",
    "Transformer",
    "Tuple",
]



from typing import Iterable

from xformula.syntax.grammar.ebnf._flat_join import flat_join

__all__ = [
    "concat",
]


def concat(expressions: Iterable[str]) -> str:
    return flat_join(" ", expressions)



import dataclasses

from xformula.syntax.ast.nodes.abc import Operand
from xformula.syntax.ast.nodes.abc import Operation as _Operation
from xformula.syntax.ast.nodes.abc import Operator

__all__ = [
    "Operation",
]


@dataclasses.dataclass()
class Operation(_Operation):

    operator: Operator = dataclasses.field(
        kw_only=True,
        default_factory=Operator,
    )

    operands: list[Operand] = dataclasses.field(
        kw_only=True,
        default_factory=list,
    )



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.ast.nodes.abc import SimpleExpression
from xformula.syntax.core.features.containers.ast.nodes import List as ListNode
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "List",
]


class List(
    NonTerminal[ListNode],
):
    class Meta:

        atomic = True

        tags = {
            non_terminal("Primary"): 3000,
        }

    def build_grammar(self) -> str:
        define = self.ebnf.define
        use_template = self.ebnf.use_template
        op = self.ebnf.op
        _non_terminal = self.ebnf.non_terminal

        return define(
            use_template(
                "WrapZeroOrMoreOf",
                op("["),
                op("]"),
                _non_terminal("SimpleExpression"),
            ),
        )

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[SimpleExpression],
    ) -> ListNode:
        return ListNode(
            elements=cast(list[SimpleExpression], tree.children),
        )



from typing import TypeVar

from xformula.syntax.grammar.templates.abc import Template

__all__ = [
    "SeparatedZeroOrMoreOf",
]


T = TypeVar("T")


class SeparatedZeroOrMoreOf(
    Template[list[T]],
):
    class Meta:

        arguments = [
            Template.arg("sep"),
            Template.arg("rule"),
        ]

    def build_grammar(self) -> str:
        arg = self.ebnf.arg
        define = self.ebnf.define
        optional = self.ebnf.optional
        use_template = self.ebnf.use_template
        suite = self.ebnf.suite

        return suite(
            define(
                optional(
                    use_template("SeparatedZeroOrMoreOf", arg("sep"), arg("rule")),
                    arg("sep"),
                ),
                arg("rule"),
            ),
        )



import dataclasses
from abc import ABCMeta
from typing import Any, Generic, NoReturn, Optional, TypeVar, cast

from xformula.arch.meta.options import Options

__all__ = [
    "Meta",
]


C = TypeVar("C", bound="Meta")

O = TypeVar("O", bound=Options)


class Meta(
    ABCMeta,
    Generic[O],
):

    options_class: type[O] = cast(type[O], Options)

    options: O

    __XF__META_ATTNAME__: str = "Meta"

    def __new__(
        mcs: type[C],
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> C:
        meta_attname = mcs.get_meta_attname(
            name,
            bases,
            namespace,
            **kwargs,
        )
        meta = mcs.get_meta(
            name,
            bases,
            namespace,
            **kwargs,
        )

        options_kwargs = mcs.get_super_options_kwargs(
            name,
            bases,
            namespace,
            **kwargs,
        )

        options_kwargs["name"] = name

        mcs.populate_options_kwargs(
            name,
            bases,
            namespace,
            meta,
            options_kwargs,
            **kwargs,
        )

        if not options_kwargs.get("name"):
            options_kwargs["name"] = name

        options_class = mcs.get_options_class()
        options = options_class(**options_kwargs)

        mcs.validate_options(
            name,
            bases,
            namespace,
            options,
            **kwargs,
        )

        namespace["options"] = options
        if not options.abstract:
            namespace.pop(meta_attname, None)

        generated_type = super(Meta, mcs).__new__(
            mcs,
            name,
            bases,
            namespace,
            **kwargs,
        )

        mcs.post_generate(
            name,
            bases,
            namespace,
            meta,
            generated_type,
            **kwargs,
        )

        return generated_type

    @classmethod
    def get_meta_attname(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> str:
        return mcs.__XF__META_ATTNAME__

    @classmethod
    def get_meta(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> Any | None:
        return namespace.get(
            mcs.get_meta_attname(
                name,
                bases,
                namespace,
                **kwargs,
            ),
            None,
        )

    @classmethod
    def get_options_class(mcs) -> type[O]:
        return mcs.options_class

    @classmethod
    def get_excluded_super_options_attnames(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> set[str]:
        return {"abstract", "name"}

    @classmethod
    def get_super_options_kwargs(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> dict[str, Any]:
        options_kwargs: dict[str, Any] = dict()
        excluded_att_names = mcs.get_excluded_super_options_attnames(
            name,
            bases,
            namespace,
            **kwargs,
        )
        for super_type in bases[::-1]:
            super_options = cast(Optional[O], getattr(super_type, "options", None))
            if super_options is None:
                continue
            for field in dataclasses.fields(mcs.options_class):
                if field.name in excluded_att_names or field.name in options_kwargs:
                    continue
                if hasattr(super_options, field.name):
                    options_kwargs[field.name] = getattr(super_options, field.name)
        return options_kwargs

    @classmethod
    def populate_options_kwargs(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        meta: Any | None,
        options_kwargs: dict[str, Any],
        **kwargs: Any,
    ) -> None:
        if meta is None:
            return
        for field in dataclasses.fields(mcs.options_class):
            if hasattr(meta, field.name):
                options_kwargs[field.name] = getattr(meta, field.name)

    @classmethod
    def post_generate(
        mcs: type[C],
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        meta: Any | None,
        generated_type: C,
        **kwargs: Any,
    ) -> None:
        return None

    @classmethod
    def validate_options(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        options: O,
        **kwargs: Any,
    ) -> NoReturn | None:
        return None



from typing import cast

from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.core.features.literals.ast.nodes import Str as StrNode
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.non_terminals.abc import NonTerminal
from xformula.syntax.parser.trees.abc import ParseTree

__all__ = [
    "Str",
]


class Str(
    NonTerminal[StrNode],
):
    class Meta:

        definition_name = "Str"

        atomic = True

        tags = {
            non_terminal("Literal"): -6000,
        }

    def build_grammar(self) -> str:
        return self.ebnf.define_tagged_alternation()

    def transform_parse_tree(
        self,
        runtime_context: RuntimeContext,
        tree: ParseTree[str],
    ) -> StrNode:
        value = cast(str, tree.children[0])
        return StrNode(
            value=value,
        )



from typing import TYPE_CHECKING

from xformula.arch.meta import Configurable
from xformula.syntax.core.features.abc.feature_meta import FeatureMeta

if TYPE_CHECKING:
    from xformula.syntax.core.context.abc.syntax_context import SyntaxContext
    from xformula.syntax.grammar.directives.abc.directive import Directive
    from xformula.syntax.grammar.non_terminals.abc.non_terminal import NonTerminal
    from xformula.syntax.grammar.templates.abc.template import Template
    from xformula.syntax.grammar.terminals.abc.terminal import Terminal

__all__ = [
    "Feature",
]


class Feature(
    Configurable,
    metaclass=FeatureMeta,
):

    _context: "SyntaxContext"

    _directive_types: list[type["Directive"]]

    _non_terminal_types: list[type["NonTerminal"]]

    _template_types: list[type["Template"]]

    _terminal_types: list[type["Terminal"]]

    class Meta:
        abstract = True

    def __init__(
        self,
        context: "SyntaxContext",
    ) -> None:
        self._context = context
        self._directive_types = []
        self._non_terminal_types = []
        self._template_types = []
        self._terminal_types = []

    @property
    def context(self) -> "SyntaxContext":
        return self._context

    @property
    def directive_types(self) -> list[type["Directive"]]:
        return self._directive_types

    @property
    def non_terminal_types(self) -> list[type["NonTerminal"]]:
        return self._non_terminal_types

    @property
    def template_types(self) -> list[type["Template"]]:
        return self._template_types

    @property
    def terminal_types(self) -> list[type["Terminal"]]:
        return self._terminal_types

    def pre_setup(self) -> None:
        ...

    def setup(self) -> None:
        ...

    def post_setup(self) -> None:
        ...

    def on_ready(self) -> None:
        ...



from typing import Any, TypeVar

from xformula.arch.meta import Meta
from xformula.syntax.core.features.abc.feature_options import FeatureOptions

__all__ = [
    "FeatureMeta",
]


C = TypeVar("C", bound="FeatureMeta")


class FeatureMeta(
    Meta[FeatureOptions],
):

    options_class = FeatureOptions

    options: FeatureOptions

    @classmethod
    def get_excluded_super_options_attnames(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> set[str]:
        excluded = super(FeatureMeta, mcs).get_excluded_super_options_attnames(
            name,
            bases,
            namespace,
            **kwargs,
        )
        excluded.add("fqn")
        return excluded

    @classmethod
    def post_generate(
        mcs: type[C],
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        meta: Any | None,
        generated_type: C,
        **kwargs: Any,
    ) -> None:
        if not generated_type.options.fqn:
            generated_type.options.fqn = ".".join(
                [
                    generated_type.__module__,
                    generated_type.__qualname__,
                ],
            )



from xformula.syntax.core.features.operations.definitions.non_terminals.operand import (
    Operand,
)
from xformula.syntax.core.features.operations.definitions.non_terminals.operation import (
    Operation,
)

__all__ = [
    "Operand",
    "Operation",
]



__all__ = [
    "one_or_more_of",
]


def one_or_more_of(*expressions: str) -> str:
    inner = " ".join(expressions)
    wrap = len(expressions) > 1 or " " in inner
    if wrap:
        return f"({inner})+"
    return f"{inner}+"



from xformula.syntax.ast.nodes.abc.context import Context
from xformula.syntax.ast.nodes.abc.node import Node

__all__ = [
    "HasContext",
]


class HasContext(Node):

    context: Context

    class Meta(Node.Meta):

        abstract = True

        has_context_attribute = True



from xformula.arch.enumeration import Iota
from xformula.syntax.ast.nodes.abc import Associativity, Placement, SymbolType
from xformula.syntax.core.operations.abc import OperatorPrecedence

__all__ = [
    "DEFAULT_OPERATOR_PRECEDENCES",
]


__xf__operator_precedence_iota__ = Iota(
    value=0,
    step=1000,
)


DEFAULT_OPERATOR_PRECEDENCES: list[OperatorPrecedence] = [
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "**"),
        ],
    ),
    OperatorPrecedence(
        arity=1,
        associativity=Associativity.RIGHT_TO_LEFT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.LEFT,
        symbols=[
            (SymbolType.OPERATOR, "+"),
        ],
    ),
    OperatorPrecedence(
        arity=1,
        associativity=Associativity.RIGHT_TO_LEFT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.LEFT,
        symbols=[
            (SymbolType.OPERATOR, "-"),
        ],
    ),
    OperatorPrecedence(
        arity=1,
        associativity=Associativity.RIGHT_TO_LEFT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.LEFT,
        symbols=[
            (SymbolType.OPERATOR, "!"),
        ],
    ),
    OperatorPrecedence(
        arity=1,
        associativity=Associativity.RIGHT_TO_LEFT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.LEFT,
        symbols=[
            (SymbolType.OPERATOR, "~"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "*"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "@"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "/"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "//"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "%"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "+"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "-"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "<<"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, ">>"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "&"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "^"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "|"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "not in"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "!in"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.KEYWORD, "in"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.KEYWORD, "is not"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "!is"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.KEYWORD, "is"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "!<"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "!<="),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "<="),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "<"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "!>"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "!>="),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, ">="),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, ">"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "!="),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "=="),
        ],
    ),
    OperatorPrecedence(
        arity=3,
        associativity=Associativity.RIGHT_TO_LEFT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.LEFT,
        symbols=[
            (SymbolType.OPERATOR, "?"),
            (SymbolType.OPERATOR, ":"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "??"),
        ],
    ),
    OperatorPrecedence(
        arity=1,
        associativity=Associativity.RIGHT_TO_LEFT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.LEFT,
        symbols=[
            (SymbolType.KEYWORD, "not"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "&&"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.OPERATOR, "||"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.KEYWORD, "and"),
        ],
    ),
    OperatorPrecedence(
        arity=2,
        associativity=Associativity.LEFT_TO_RIGHT,
        level=__xf__operator_precedence_iota__(),
        placement=Placement.RIGHT,
        symbols=[
            (SymbolType.KEYWORD, "or"),
        ],
    ),
]


del __xf__operator_precedence_iota__



from xformula.runtime.core.context.abc import RuntimeContext
from xformula.syntax.grammar.ebnf import non_terminal
from xformula.syntax.grammar.terminals.abc import Terminal
from xformula.syntax.lexer.tokens.abc import Token

__all__ = [
    "HEXADECIMAL_INT",
]


class HEXADECIMAL_INT(  # NOSONAR
    Terminal[int],
):
    class Meta:

        tags = {
            non_terminal("Int"): -3000,
        }

    def build_grammar(self) -> str:
        define = self.ebnf.define
        regex = self.ebnf.regex

        any_char_of = self.regex.any_char_of
        chars_between = self.regex.chars_between
        flat_one_or_more_of = self.regex.flat_one_or_more_of
        flat_optional = self.regex.flat_optional
        non_capturing_group = self.regex.non_capturing_group

        return define(
            regex(
                "0",
                any_char_of("x", "X"),
                flat_one_or_more_of(
                    non_capturing_group(
                        flat_optional("_"),
                        any_char_of(
                            chars_between("0", "9"),
                            chars_between("a", "f"),
                            chars_between("A", "F"),
                        ),
                    ),
                ),
            ),
        )

    def transform_token(
        self,
        runtime_context: RuntimeContext,
        token: Token,
    ) -> int:
        return int(token.value, base=16)



from typing import TYPE_CHECKING

from xformula.syntax.grammar.ebnf._directive import directive
from xformula.syntax.grammar.ebnf._non_terminal import non_terminal
from xformula.syntax.grammar.ebnf._template import template
from xformula.syntax.grammar.ebnf._terminal import terminal

if TYPE_CHECKING:
    from xformula.syntax.grammar.definitions.abc.definition_type import DefinitionType

__all__ = [
    "normalize_name_by_type",
]


def normalize_name_by_type(
    definition_type: "DefinitionType",
    name: str,
) -> str:
    from xformula.syntax.grammar.definitions.abc.definition_type import DefinitionType

    dispatcher = {
        DefinitionType.DIRECTIVE: directive,
        DefinitionType.TERMINAL: terminal,
        DefinitionType.TEMPLATE: template,
        DefinitionType.NON_TERMINAL: non_terminal,
    }

    normalize = dispatcher[definition_type]
    return normalize(name)



from xformula.syntax.core.features.abc import Feature

__all__ = [
    "BaseFeature",
]


class BaseFeature(Feature):
    class Meta:
        fqn = "xformula.core.Base"

    def setup(self) -> None:
        from xformula.syntax.core.features.base.definitions.directives import (
            SkipWhitespace,
        )
        from xformula.syntax.core.features.base.definitions.terminals import WHITESPACE

        self.directive_types.append(SkipWhitespace)
        self.terminal_types.append(WHITESPACE)



from typing import Iterable, cast

from xformula.syntax.grammar.regex._escape_char import escape_char

__all__ = [
    "any_char_of",
]


def any_char_of(*chars: str) -> str:
    inner = "".join(
        map(
            escape_char,
            cast(
                Iterable[str],
                filter(
                    len,
                    chars,
                ),
            ),
        ),
    )
    if not inner:
        return ""
    return f"[{inner}]"



from typing import TypeVar

from xformula.syntax.ast.nodes.abc.has_elements import HasElements
from xformula.syntax.ast.nodes.abc.node import Node
from xformula.syntax.ast.nodes.abc.term import Term

__all__ = [
    "Container",
]


T = TypeVar("T")


class Container(
    Term,
    HasElements[list[T]],
    Node,
):
    class Meta(
        Term.Meta,
        HasElements.Meta,
        Node.Meta,
    ):

        abstract = True

        is_container = True



import dataclasses

from xformula.syntax.ast.nodes.abc import Symbol as _Symbol
from xformula.syntax.ast.nodes.abc import SymbolType

__all__ = [
    "Symbol",
]


@dataclasses.dataclass()
class Symbol(_Symbol):

    type: SymbolType = dataclasses.field(
        kw_only=True,
        default=SymbolType.UNDEFINED,
    )

    value: str = dataclasses.field(
        kw_only=True,
        default=str(),
    )
